
#include <algorithm> // for std::sort
#include "KDTreeScene.h"
#include "TriangleMesh.h" // for intersectRayAABB()
                          // TODO: move that function somewhere else

// The KD-Tree depth can NEVER exceed this amount. Although in practical
// cases the maximum depth of the tree will be much less.
#define KDTREE_MAX_DEPTH   64


const float KDTreeScene::intersectionCost = 80.0f;
const float KDTreeScene::traversalCost    = 1.0f;
const float KDTreeScene::emptyBonus       = 0.5f;
const int   KDTreeScene::maxPrimsInLeaf   = 1;
const int   KDTreeScene::maxBadRefines    = 3;


KDTreeScene::KDTreeScene():
     aabb( Point3(0.0f, 0.0f, 0.0f), Point3(0.0f, 0.0f, 0.0f) ),
     root(NULL) {

}


KDTreeScene::~KDTreeScene() {
  clear();
}


void KDTreeScene::clear() {
  if ( root != NULL ) {
    delete root;
    root = NULL;
  }

  // delete each object?
  for ( size_t i=0; i < objects.size(); i++ )
    if ( objects[i] )
      delete objects[i];

  objects.clear();
  
  this->Scene::clearScene();
}


void KDTreeScene::build() {
  // delete tree if it already exists
  // (don't call clear because we don't want to get rid of objects)
  if ( root ) {
    delete root;
    root = NULL;
  }

  // don't continue if there are no objects
  if ( objects.size() < 1 )
    return;

  // find AABB of entire scene
  aabb = calcAABB(objects);

  // calculate maximum depth of tree
  int maxTreeDepth = int( 8.0 + 1.3*log( double(objects.size()) ) );

  // make sure maxTreeDepth is not too big
  if ( maxTreeDepth > KDTREE_MAX_DEPTH )
    maxTreeDepth = KDTREE_MAX_DEPTH;

  // create vector containing all object indices in root, this is required
  // by the KDTreeNode constructor which recursively builds the KD-Tree.
  // KDTreeNode() will internally create new vectors of object indices
  // to indicate which objects intersect a node's parent.
  std::vector<size_t> rootObjectIndices( objects.size() );
  for ( size_t i=0; i < objects.size(); i++ )
    rootObjectIndices[i] = i;

  // store AABB of each object so they don't have to be recomputed during
  // the construction of the KD-Tree
  std::vector<AABB> objectAABBs( objects.size() );
  for ( size_t i=0; i < objects.size(); i++ )
    objectAABBs[i] = objects[i]->getAABB();

  // call KDTreeNode contructor which will recursively build the KD-Tree
  root = new KDTreeNode(aabb,
                        objects,
                        maxPrimsInLeaf,
                        rootObjectIndices,
                        objectAABBs,
                        0,
                        maxTreeDepth);

  if ( !root )
    throw RTException("KDTreeScene::build(): unable to allocate root");
}


void KDTreeScene::insertObject(SceneObject *object) {
  if ( object != NULL )
    objects.push_back(object);
}


void KDTreeScene::insertLight(const Light &light) {
  lights.push_back(light);
}


// utility function to find the AABB of a vector of objects
AABB KDTreeScene::calcAABB(std::vector<SceneObject *> &objs) {
  Point3 minP( RT_FLT_MAX,  RT_FLT_MAX,  RT_FLT_MAX ),
         maxP(-RT_FLT_MAX, -RT_FLT_MAX, -RT_FLT_MAX );

  for ( size_t i=0; i < objs.size(); i++ ) {
    AABB objectAABB = objs[i]->getAABB();

    Point3 objMin = objectAABB.getMin(),
           objMax = objectAABB.getMax();

    for ( int j=0; j < 3; j++ ) {
      if ( objMin[j] < minP[j] )
        minP[j] = objMin[j];
      if ( objMax[j] > maxP[j] )
        maxP[j] = objMax[j];
    } // for each axis
  }

  return AABB(minP, maxP);
}


SceneObject *KDTreeScene::findNearestHit(const Ray &ray,
                                         HitInfo &hitInfoOut) {
  struct KDTreeNodeQueueEl {
    KDTreeNode *node;
    float tmin, tmax;
  };
  KDTreeNodeQueueEl nodeQueue[KDTREE_MAX_DEPTH];
  int nodeQueuePos = 0;

  // pbrt: compute initial parametric range of ray inside kd-tree extent
  float tmin = 0.0f,
        tmax = 0.0f;
  if ( !intersectRayAABB(ray, aabb, &tmin, &tmax) )
    return NULL; // ray does not intersect bounding box of scene

  // store inverse of ray direction so multiplies can be used instead of
  // divides in loop
  const Vector3 &rayDir    = ray.getDir();
  const Point3  &rayOrigin = ray.getOrigin();
  Vector3 invDir(1.0f/rayDir.m[0],
                 1.0f/rayDir.m[1],
                 1.0f/rayDir.m[2]);

  SceneObject *nearestHit = NULL;
  HitInfo nearestHitInfo, objHitInfo;
  nearestHitInfo.t = objHitInfo.t = RT_FLT_MAX;

  // pbrt: Traverse kd-tree nodes in order for ray
  //bool hit = false;
  KDTreeNode *node = root;
  while ( node != NULL ) {
    // pbrt: bail out if we found a hit closer than the current node
    if ( nearestHitInfo.t < tmin )
      break;

    unsigned char axis = node->getAttrs();
    
    if ( axis != 0x03 ) {
      // pbrt: process kd-tree interior node
      const KDTreeNode::InternalNodeData &internalNodeData = node->getInternalNodeData();

      // pbrt: compute parametric distance along ray to split plane
      float tplane = (internalNodeData.planeOffset - rayOrigin.m[axis])*invDir[axis];

      // pbrt: get node children pointers for ray
      KDTreeNode *firstChild,
                 *secondChild;
      if ( rayOrigin.m[axis] <= internalNodeData.planeOffset ) {
        firstChild  = internalNodeData.left;
        secondChild = internalNodeData.right;
      }
      else {
        firstChild  = internalNodeData.right;
        secondChild = internalNodeData.left;
      }

      // pbrt: advance to next child node, possibly enqueue other child
      if ( tplane > tmax || tplane < 0.0f )
        node = firstChild;
      else if ( tplane < tmin )
        node = secondChild;
      else {
        // pbrt: enqueue second child in todo list
        nodeQueue[nodeQueuePos].node = secondChild;
        nodeQueue[nodeQueuePos].tmin = tplane;
        nodeQueue[nodeQueuePos].tmax = tmax;
        nodeQueuePos++;
        node = firstChild;
        tmax = tplane;
      }
    } // if ( internal node )
    else {
      // pbrt: check for intersections inside leaf node
      const KDTreeNode::LeafNodeData &leafNodeData = node->getLeafNodeData();
      for ( int i=0; i < leafNodeData.numObjects; i++ ) {
        SceneObject *obj = leafNodeData.objects[i];
        // pbrt: check one primitive inside leaf node
        if ( obj->intersectRay(ray, objHitInfo) ) {
          // is this the closest hit?
          if ( objHitInfo.t < nearestHitInfo.t && objHitInfo.t > 0.0f ) {
            nearestHitInfo = objHitInfo;
            nearestHit = obj;
          }
        } // if ( ray intersects object )
      } // for ( objects in leaf node )

      // pbrt: grab next node to process from todo list
      if ( nodeQueuePos > 0 ) {
        nodeQueuePos--;
        node = nodeQueue[nodeQueuePos].node;
        tmin = nodeQueue[nodeQueuePos].tmin;
        tmax = nodeQueue[nodeQueuePos].tmax;
      }
      else
        break;
    } // else ( leaf node )
  } // while ( node != NULL )

  if ( nearestHit )
    hitInfoOut = nearestHitInfo;

  return nearestHit;
}


// returns after first hit that is within range specified by maxT
bool KDTreeScene::occlusionTest(const Ray &ray, float maxT) const {
    struct KDTreeNodeQueueEl {
    KDTreeNode *node;
    float tmin, tmax;
  };
  KDTreeNodeQueueEl nodeQueue[KDTREE_MAX_DEPTH];
  int nodeQueuePos = 0;

  // pbrt: compute initial parametric range of ray inside kd-tree extent
  float tmin = 0.0f,
        tmax = 0.0f;
  if ( !intersectRayAABB(ray, aabb, &tmin, &tmax) )
    return false; // ray does not intersect bounding box of scene

  // store inverse of ray direction so multiplies can be used instead of
  // divides in loop
  const Vector3 &rayDir    = ray.getDir();
  const Point3  &rayOrigin = ray.getOrigin();
  Vector3 invDir(1.0f/rayDir.m[0],
                 1.0f/rayDir.m[1],
                 1.0f/rayDir.m[2]);

  // pbrt: Traverse kd-tree nodes in order for ray
  //bool hit = false;
  KDTreeNode *node = root;
  while ( node != NULL ) {
    // TODO: maybe store t of nearest hit so can break here again
    // pbrt: bail out if we found a hit closer than the current node
    //if ( nearestHitInfo.t < tmin )
    //  break;

    unsigned char axis = node->getAttrs();
    
    if ( axis != 0x03 ) {
      // pbrt: process kd-tree interior node
      const KDTreeNode::InternalNodeData &internalNodeData = node->getInternalNodeData();

      // pbrt: compute parametric distance along ray to split plane
      float tplane = (internalNodeData.planeOffset - rayOrigin.m[axis])*invDir[axis];

      // pbrt: get node children pointers for ray
      KDTreeNode *firstChild,
                 *secondChild;
      if ( rayOrigin.m[axis] <= internalNodeData.planeOffset ) {
        firstChild  = internalNodeData.left;
        secondChild = internalNodeData.right;
      }
      else {
        firstChild  = internalNodeData.right;
        secondChild = internalNodeData.left;
      }

      // pbrt: advance to next child node, possibly enqueue other child
      if ( tplane > tmax || tplane < 0.0f )
        node = firstChild;
      else if ( tplane < tmin )
        node = secondChild;
      else {
        // pbrt: enqueue second child in todo list
        nodeQueue[nodeQueuePos].node = secondChild;
        nodeQueue[nodeQueuePos].tmin = tplane;
        nodeQueue[nodeQueuePos].tmax = tmax;
        nodeQueuePos++;
        node = firstChild;
        tmax = tplane;
      }
    } // if ( internal node )
    else {
      // pbrt: check for intersections inside leaf node
      const KDTreeNode::LeafNodeData &leafNodeData = node->getLeafNodeData();
      for ( int i=0; i < leafNodeData.numObjects; i++ ) {
        if ( leafNodeData.objects[i]->occlusionTest(ray, maxT) )
          return true;
      } // for ( objects in leaf node )

      // pbrt: grab next node to process from todo list
      if ( nodeQueuePos > 0 ) {
        nodeQueuePos--;
        node = nodeQueue[nodeQueuePos].node;
        tmin = nodeQueue[nodeQueuePos].tmin;
        tmax = nodeQueue[nodeQueuePos].tmax;
      }
      else
        break;
    } // else ( leaf node )
  } // while ( node != NULL )

  return false;
}


RGBColor<float> KDTreeScene::lightContrib(const Ray &ray,
                                          const Point3 &p,
                                          const Vector3 &n,
                                          int matID,
                                          const Point2 &tc) {
  RGBColor<float> color(0.0f, 0.0f, 0.0f);
  RGBColor<float> texSampleColor(1.0f, 1.0f, 1.0f);
  const Material &material = materials[matID];

  // if this material has a texture, sample from it
  if ( material.texture != NULL )
    texSampleColor = material.texture->sampleBilinear(tc[0], tc[1]);

  // add contribution from each light
  for ( size_t i=0; i < lights.size(); i++ )
    color += singleLightContrib(lights[i], p, ray, n, material);

  return texSampleColor*color;
}


KDTreeScene::KDTreeNode::KDTreeNode(const AABB &nodeAABB,
                                    const std::vector<SceneObject *> &sceneObjects,
                                    size_t maxLeafPrims,
                                    const std::vector<size_t> &objectsInNode,
                                    const std::vector<AABB> &objectAABBs,
                                    int badRefines,
                                    int depth) {
  // zero out self before doing anything
  // just memset the larger of the two node structures to 0, and
  // set attrs to zero
  attrs = 0;
  memset( &internalNodeData, 0, sizeof(internalNodeData) );

  // termination criteria met?
  if ( objectsInNode.size() < maxLeafPrims || depth < 1 ) {
    // create a leaf node
    createLeafNode(sceneObjects, objectsInNode);
  }
  else {
    // try to create an internal node
    createInternalNode(nodeAABB,
                       sceneObjects,
                       objectsInNode,
                       objectAABBs,
                       maxLeafPrims,
                       badRefines,
                       depth);
  }
}


KDTreeScene::KDTreeNode::~KDTreeNode() {
  // is this an internal node, or a leaf node?
  if ( isLeaf() ) {
    // leaf node
    if ( leafNodeData.objects )
      delete [] leafNodeData.objects;
  }
  else {
    // internal node
    if ( internalNodeData.left )
      delete internalNodeData.left;
    if ( internalNodeData.right )
      delete internalNodeData.right;
  }
}


void KDTreeScene::KDTreeNode::createLeafNode(const std::vector<SceneObject *> &sceneObjects,
                                             const std::vector<size_t> &objectsInNode) {
  // set attrs to indicate this is a leaf
  attrs = 0x03;

  leafNodeData.numObjects = int( objectsInNode.size() );
  leafNodeData.objects = new SceneObject *[leafNodeData.numObjects];

  // throw an error if could not allocate storage for objects
  if ( !leafNodeData.objects )
    throw RTException("KDTreeScene::KDTreeNode::createLeaf(): unable to allocate storage for objects");

  for ( int i=0; i < leafNodeData.numObjects; i++ )
    leafNodeData.objects[i] = sceneObjects[ objectsInNode[i] ];
}


void KDTreeScene::KDTreeNode::createInternalNode(const AABB &nodeAABB,
                                                 const std::vector<SceneObject *> &sceneObjects,
                                                 const std::vector<size_t> &objectsInNode,
                                                 const std::vector<AABB> &objectAABBs,
                                                 size_t maxLeafPrims,
                                                 int badRefines,
                                                 int depth) {
  // choose split axis position for interior node
  unsigned char bestAxis = 3;
  size_t bestOffsetIndex = 0;
  float oldCost  = float( objectsInNode.size() )*intersectionCost,
        bestCost = RT_FLT_MAX;
        //bestCost = INFINITY;
  Vector3 d = nodeAABB.getMax() - nodeAABB.getMin();
  float totalSA = 2.0f*(d[0]*d[1] + d[0]*d[2] + d[1]*d[2]);
  float invTotalSA = 1.0f/totalSA;

  // choose which axis to split along
  unsigned char axis = d[0] > d[1] ? 0 : 1;
  if ( d[2] > d[axis] )
    axis = 2;

  std::vector<BoxEdge> boxEdges[3];

  int retries = 0;
  float nodeAABBMin,
        nodeAABBMax;
  retrySplit:

  boxEdges[axis].resize(objectsInNode.size()*2);
  for ( size_t i=0; i < objectsInNode.size(); i++ ) {
    size_t objIndex = objectsInNode[i];
    boxEdges[axis][2*i  ] = BoxEdge(objectAABBs[objIndex].getMin().m[axis], objIndex, true);
    boxEdges[axis][2*i+1] = BoxEdge(objectAABBs[objIndex].getMax().m[axis], objIndex, false);
  }

  // sort the edges
  std::sort( boxEdges[axis].begin(), boxEdges[axis].end() );

  // compute cost of all splits for axis to find best
  nodeAABBMin = nodeAABB.getMin()[axis],
  nodeAABBMax = nodeAABB.getMax()[axis];
  size_t numBelow = 0,
         numAbove = objectsInNode.size();
  for ( size_t i=0; i < 2*objectsInNode.size(); i++ ) {
    const BoxEdge &edge = boxEdges[axis][i];
    if ( false == edge.start )
      numAbove--;
    if ( edge.pos > nodeAABBMin &&
         edge.pos < nodeAABBMax ) {
      // compute cost for split at ith edge
      static const int otherAxis[3][2] = { {1, 2}, {0, 2}, {0, 1} };
      int otherAxis0 = otherAxis[axis][0],
          otherAxis1 = otherAxis[axis][1];

      // calculate surface area above and below the split
      float belowSA = 2.0f*( d[otherAxis0]*d[otherAxis1] +
                             (edge.pos - nodeAABBMin)*
                             (d[otherAxis0] + d[otherAxis1]) ),
            aboveSA = 2.0f*( d[otherAxis0]*d[otherAxis1] +
                             (nodeAABBMax - edge.pos)*
                             (d[otherAxis0] + d[otherAxis1]) );

      // compute probabilities that a random ray hitting this node will cross
      // above or below the split
      float pBelow = belowSA*invTotalSA,
            pAbove = aboveSA*invTotalSA;

      // add extra weight if empty above or below split
      float eb = ( numAbove == 0 || numBelow == 0 ) ? emptyBonus : 0.0f;

      float cost = traversalCost + intersectionCost*(1.0f - eb)*
                   (pBelow*float(numBelow) + pAbove*float(numAbove));

      // update best split if this is lowest cost so far
      if ( cost < bestCost ) {
        bestCost = cost;
        bestAxis = axis;
        bestOffsetIndex = i;
      }
    }
    if ( true == edge.start )
      numBelow++;
  } // for (edges)

  // create leaf if no good splits were found
  if ( 3 == bestAxis && retries < 2 ) {
    retries++;
    axis = (axis+1)%3; // try next axis
    goto retrySplit;
  }

  if ( bestCost > oldCost )
    badRefines++;

  if ( (bestCost > 4.0f*oldCost && objectsInNode.size() < 16) ||
       3 == bestAxis || maxBadRefines == badRefines ) {
    // create a leaf
    createLeafNode(sceneObjects, objectsInNode);
    return; // don't go any further after creating leaf
  }

  // classify primitives with respect to split
  std::vector<size_t> objects0;
  for ( size_t i=0; i < bestOffsetIndex; i++ ) {
    const BoxEdge &edge = boxEdges[bestAxis][i];
    if ( true == edge.start )
      objects0.push_back(edge.objIndex);
  }

  std::vector<size_t> objects1;
  for ( size_t i=bestOffsetIndex+1; i < 2*objectsInNode.size(); i++ ) {
    const BoxEdge &edge = boxEdges[bestAxis][i];
    if ( false == edge.start )
      objects1.push_back(edge.objIndex);
  }

  // recursively initialize child nodes
  Point3 box1Min = nodeAABB.getMin(),
         box0Max = nodeAABB.getMax();
  box0Max.m[bestAxis] = box1Min.m[bestAxis] = boxEdges[bestAxis][bestOffsetIndex].pos;
  AABB box0( nodeAABB.getMin(), box0Max ),
       box1( box1Min, nodeAABB.getMax() );

  // set attrs to best split axis to indicate that this is an internal node
  attrs = bestAxis;

  // store offset of splitting plane
  internalNodeData.planeOffset = boxEdges[bestAxis][bestOffsetIndex].pos;

  // create child nodes
  internalNodeData.left = new KDTreeNode(box0,
                                         sceneObjects,
                                         maxLeafPrims,
                                         objects0,
                                         objectAABBs,
                                         badRefines,
                                         depth-1);

  if ( !internalNodeData.left )
    throw RTException("KDTreeScene::KDTreeNode::createInternalNode(): "
                      "unable to create child node");

  internalNodeData.right = new KDTreeNode(box1,
                                          sceneObjects,
                                          maxLeafPrims,
                                          objects1,
                                          objectAABBs,
                                          badRefines,
                                          depth-1);

  if ( !internalNodeData.right )
    throw RTException("KDTreeScene::KDTreeNode::createInternalNode(): "
                      "unable to create child node");
}


KDTreeScene::BoxEdge::BoxEdge(float _pos, size_t _objIndex, bool _start):
     pos(_pos),
     objIndex(_objIndex),
     start(_start) {

}


bool KDTreeScene::BoxEdge::operator < (const BoxEdge &e) const {
  if ( pos == e.pos )
    return (true == start && false == e.start);

  return pos < e.pos;
}
