
#ifndef AABB_H
#define AABB_H

#include <iostream>
#include "Vector3.h"
#include "Point3.h"


namespace Math3D {


// an axis-aligned bounding box
template<class Real>
class AABB {
public:
  inline AABB();
  inline AABB(const Point3<Real> &_center, const Vector3<Real>& _ext);
  inline AABB(const Point3<Real>& minP, const Point3<Real>& maxP);

  inline const Point3<Real>& getCenter() const { return center; }
  inline const Vector3<Real>& getExt() const { return ext; }

  inline Point3<Real> getMin() const { return minP; }
  inline Point3<Real> getMax() const { return maxP; }

  inline bool pointInside(const Point3<Real> &p) const;

private:
  // 2 representations
  // 1) center + extents
  Point3<Real> center;
  Vector3<Real> ext;

  // 2) minP and maxP points
  Point3<Real> minP, maxP;
};


template<class Real>
inline AABB<Real>::AABB():
     center(0.0, 0.0, 0.0),
     ext(0.0, 0.0, 0.0),
     minP(0.0, 0.0, 0.0),
     maxP(0.0, 0.0, 0.0) {

}


template<class Real>
inline AABB<Real>::AABB(const Point3<Real> &_center, const Vector3<Real>& _ext):
     center(_center),
     ext(_ext),
     minP(_center - _ext),
     maxP(_center + _ext) {

}


template<class Real>
inline AABB<Real>::AABB(const Point3<Real>& _minP, const Point3<Real>& _maxP) {
  minP = _minP;
  maxP = _maxP;
  center = (minP + maxP)*Real(0.5);
  ext = maxP - center;
}


// overloaded stream insertion operator for AABB
template<class Real>
std::ostream& operator << (std::ostream &out, const AABB<Real> &aabb) {
  // write min and max points to stream
  Point3<Real> minP = aabb.getMin(),
               maxP = aabb.getMax();
  out << "(<" << minP[0] << ", " << minP[1] << ", " << minP[2] << ">" << ", "
      << "<"  << maxP[0] << ", " << maxP[1] << ", " << maxP[2] << ">" << ")";
  return out;
}


// returns true if given point is within AABB
template<class Real>
inline bool AABB<Real>::pointInside(const Point3<Real> &p) const {
  return ( p.m[0] > minP.m[0] ) &&
         ( p.m[1] > minP.m[1] ) &&
         ( p.m[2] > minP.m[2] ) &&
         ( p.m[0] < maxP.m[0] ) &&
         ( p.m[1] < maxP.m[1] ) &&
         ( p.m[2] < maxP.m[2] );
}


} // namespace Math3D


#endif
