
#include "RayTracer.h"
#include <xmmintrin.h>

#define RENDER_THREAD_TILE_SIZE  32


RayTracer::RayTracer(int w, int h, int nThreads, Scene *_scene) {
  imageWidth           = w;
  imageHeight          = h;
  numThreads           = nThreads;
  numThreadWorkTiles   = 0;
  usingFixedRegions    = false;
  currentDynamicRegion = 0;
  dynamicRegionLock    = NULL;
  scene                = _scene;
  threads              = NULL;
  viewDistance         = 1.0f;
  dynamicRegions       = NULL;
  rayDirs              = NULL;
  rotatedRayDirs       = NULL;

  // make sure both width and height are > 0
  if ( w < 1 || h < 1 )
    throw RTException("image dimension is less than 1");

  // compute view distance
  float fovY = RT_PI*0.5f; // constant for now
  viewDistance = 1/tan(fovY*0.5f);

  // allocate space for image data
  imageData = new Uint8[w*h*3];

  // was image allocated successfully?
  if ( !imageData )
    throw RTException("unable to allocate space for image");

  // zero out image
  memset(imageData, 0, w*h*3);
  
  // allocate storage for rays
  rayDirs = new Vector3[w*h];
  
  if ( !rayDirs )
    throw RTException("unable to allocate storage for ray directions");
    
  // allocate storage for rotated ray Directions
  rotatedRayDirs = new Vector3[w*h];
  
  if ( !rotatedRayDirs )
    throw RTException("unable to allocate storage for transformed "
                      "ray directions");

  float longerDim = float(imageWidth);
  if ( imageWidth < imageHeight )
    longerDim = float(imageHeight);

  // initialize ray directions    
  for (int i=0; i < w; i++) {
    for (int j=0; j < h; j++) {
      float dI = float(i)/longerDim - 0.5f,
            dJ = float(j)/longerDim - 0.5f;

      Vector3 dir = Vector3( dI, dJ, -viewDistance );
      dir.normalize();
      rotatedRayDirs[j*w + i] = rayDirs[j*w + i] = dir;
    }
  }
  
  if ( nThreads > 1 ) {
    threads = new SDL_Thread*[nThreads];

    if ( !threads )
      throw RTException("unable to allocate storage for thread information");

    if ( !usingFixedRegions ) {
      // calculate how many thread work tiles are needed
      int numTilesX = w/RENDER_THREAD_TILE_SIZE +
                      (w%RENDER_THREAD_TILE_SIZE == 0 ? 0 : 1),
          numTilesY = h/RENDER_THREAD_TILE_SIZE +
                      (h%RENDER_THREAD_TILE_SIZE == 0 ? 0 : 1);
      numThreadWorkTiles = numTilesX*numTilesY;

      // allocate render thread work stack
      dynamicRegions = new DynamicRegion[numThreadWorkTiles];

      if ( !dynamicRegions )
        throw RTException("unable to allocate storage for thread work tiles");

      // create a lock for thread work tiles
      dynamicRegionLock = SDL_CreateMutex();

      if ( !dynamicRegionLock )
        throw RTException("could not create lock for thread work tiles");

      // set work regions
      for ( int x=0; x < numTilesX; x++ ) {
        for ( int y=0; y < numTilesY; y++ ) {
          DynamicRegion &tile = dynamicRegions[y*numTilesX + x];
          tile.x0 = x*RENDER_THREAD_TILE_SIZE;
          tile.y0 = y*RENDER_THREAD_TILE_SIZE;
          tile.x1 = tile.x0 + RENDER_THREAD_TILE_SIZE;
          tile.y1 = tile.y0 + RENDER_THREAD_TILE_SIZE;

          if ( tile.x1 >= w )
            tile.x1 = w-1;
          if ( tile.y1 >= h )
            tile.y1 = h-1;
        } // for (tiles y)
      } // for (tiles x)
    } // if (using dynamic regions)
  } // if (number of threads > 1)
} // RayTracer::RayTracer()


RayTracer::~RayTracer() {
  if ( imageData )
    delete [] imageData;
    
  if ( rayDirs )
    delete [] rayDirs;

  if ( rotatedRayDirs )
    delete [] rotatedRayDirs;

  if ( threads )
    delete [] threads;

  if ( dynamicRegions )
    delete [] dynamicRegions;

  if ( dynamicRegionLock )
    SDL_DestroyMutex(dynamicRegionLock);
} // RayTracer::~RayTracer()


// intended to be overridden by subclasses
void RayTracer::mainLoop() {
  render(0, 0, imageWidth, imageHeight);
} // RayTracer::mainLoop()


const RayTracer::DynamicRegion* RayTracer::getNextDynamicRegion() {
  SDL_mutexP(dynamicRegionLock);
  
  // if there are no more regions left, then return NULL
  if ( currentDynamicRegion >= numThreadWorkTiles ) {
    SDL_mutexV(dynamicRegionLock);
    return NULL;
  }
  
  const DynamicRegion *region = &dynamicRegions[currentDynamicRegion];
  currentDynamicRegion++;
  SDL_mutexV(dynamicRegionLock);

  return region;
} // RayTracer::getNextDynamicRegion()


RGBColor<float> RayTracer::traceRay(const Ray &ray,
                                    float refrIndex,
                                    int depth, int maxDepth) {
  const Material *materials = scene->getMaterials();
  RGBColor<float> color(0.0f, 0.0f, 0.0f);

  // find nearest object that ray hits
  HitInfo hitInfo;
  hitInfo.t = RT_FLT_MAX;
  SceneObject *nearestHit = scene->findNearestHit(ray, hitInfo);

  // did the ray hit anything at all?
  if ( nearestHit != NULL && hitInfo.t >= 0.0f ) {
    Point3 hitPos = ray.getOrigin() + ray.getDir()*hitInfo.t;

    // get material properties and texture coordinate at hitpoint
    Point2 tc(0.0f, 0.0f);
    
    // get material ID of object
    // (currently only one material for each object)
    int matID = nearestHit->getMatID();
    
    // TODO: perhaps matID must be checked for validity somewhere
    
    // get normal and texture coordinates at hitpoint
    Vector3 normal;
    nearestHit->getNormalAndTCAt(hitPos,
                                 hitInfo.texInfo,
                                 materials[matID],
                                 tc,
                                 normal);

    color = scene->lightContrib(ray, hitPos, normal, matID, tc);

    // if we have not reached maximum depth of recursion,
    // cast reflected and refracted rays
    if ( depth < maxDepth ) {
      float dot = ray.getDir()*normal;

      // reflected ray
      if ( materials[matID].reflectivity > 0.0f ) {
        // compute direction of reflected ray
        Vector3 reflDir = ray.getDir() - normal*( dot*2.0f );

        // cast reflected ray
        color += traceRay(Ray(hitPos + normal*RT_EPSILON, reflDir),
                          refrIndex, depth+1,
                          maxDepth)*materials[matID].reflectivity;
      }

      // refracted ray
      if ( materials[matID].refrIndex > 0.0f ) {
        float angle = -dot;
        float ratio = refrIndex/materials[matID].refrIndex;
        float diff  = 1.0f - ratio*ratio*(1.0f - angle*angle);
        if ( diff > 0.0f ) {
          // compute direction of refracted ray
          Vector3 refrDir = ray.getDir()*ratio + 
                            normal*( ratio*angle - sqrt(diff) )*-1.0f;
          
          // cast refracted ray
          color += traceRay(Ray(hitPos - normal*RT_EPSILON, refrDir),
                            materials[matID].refrIndex, depth+1, maxDepth);
        }
      } // if (material is refractive)

    } // if ( max depth not reached )
  }
  else {
    // the ray did not hit anything, so add a contribution from the skybox
    // if there is one
    if ( scene->getSkyBox() )
      scene->getSkyBox()->intersectRay(ray, color);
  }

  return color;
}


void RayTracer::render(int x, int y, int w, int h) {
  // if using multiple threads, then partition image and assign
  // a region to each thread
  if ( numThreads > 1 ) {
    if ( usingFixedRegions ) {
      int regionHeight = h/numThreads;
      int remainder    = h%numThreads;

      // spawn threads
      RenderThreadArgs args;
      args.rayTracer = this;
      args.x = x;
      args.w = w;
      args.h = regionHeight;

      int yPos = y + remainder;
      for ( int i=0; i < numThreads; i++ ) {
        args.y = yPos + i*regionHeight;
        threads[i] = SDL_CreateThread(renderThread, &args);

        if ( NULL == threads[i] )
          throw RTException("unable to create rendering thread");
      }

      // wait for all threads to finish
      for ( int i=0; i < numThreads; i++ )
        SDL_WaitThread(threads[i], NULL);
    } // if (using fixed regions)
    else {
      // using dynamic regions

      // reset dynamic region index
      currentDynamicRegion = 0;

      // create threads
      for ( int i=0; i < numThreads; i++ ) {
        threads[i] = SDL_CreateThread(renderThreadDynamicRegions, this);

        if ( NULL == threads[i] )
          throw RTException("unable to create dynamic rendering thread");
      }

      // wait for all threads to finish
      for ( int i=0; i < numThreads; i++ )
        SDL_WaitThread(threads[i], NULL);
    } // else (using dynamic regions)
  }
  else {
    // not using multiple threads
    renderInternal(x, y, w, h);
  }
} // RayTracer::render()


int RayTracer::renderThread(void *data) {
  RenderThreadArgs *args = (RenderThreadArgs *)data;

  args->rayTracer->renderInternal(args->x, args->y, args->w, args->h);

  return 1;
} // RayTracer::renderThread()


int RayTracer::renderThreadDynamicRegions(void *rayTracerP) {
  RayTracer *rayTracer = (RayTracer *)rayTracerP;

  const DynamicRegion *region = rayTracer->getNextDynamicRegion();
  while ( region ) {
    rayTracer->renderInternal(region->x0,
                              region->y0,
                              region->x1 - region->x0 + 1,
                              region->y1 - region->y0 + 1);

    region = rayTracer->getNextDynamicRegion();
  }

  return 1;
} // RayTracer::renderThreadDynamicRegions()


void RayTracer::renderInternal(int x, int y, int w, int h) {
  // sanity check
  if ( NULL == scene  || NULL == imageData ||
       x > imageWidth || y > imageHeight )
    return;

  float longerDim = float(imageWidth);
  if ( imageWidth < imageHeight )
    longerDim = float(imageHeight);

  int xEnd = x + w,
      yEnd = y + h;

  for (int j=y; j < yEnd; j++) {
    int jMulW = j*imageWidth;
    for (int i=x; i < xEnd; i++) {
      Ray ray;
      ray.setOrigin( camera.getPos() );
      ray.setDirDontNormalize(rotatedRayDirs[jMulW + i]);

      RGBColor<float> color = traceRay(ray);

      // clamp color to range [0, 1.0]
      color.clamp(0.0f, 1.0f);

      int index = (j*imageWidth + i)*3;
      
      // slow ftoi() method to convert colors
      Uint8 bColor[3] = { Uint8(color.c[0]*255.0f),
                          Uint8(color.c[1]*255.0f),
                          Uint8(color.c[2]*255.0f) };

      imageData[index  ] = bColor[0];
      imageData[index+1] = bColor[1];
      imageData[index+2] = bColor[2];
    }
  }
  
} // RayTracer::renderInternal() (no packet tracing)


// TODO: get rid of this.  eliminate public access to camera and just
//       write methods to update camera position and direction
void RayTracer::updateRayDirs() {
  Matrix3 rotX, rotY, rotMat;
  rotX.rotx( camera.getXRot() );
  rotY.roty( camera.getYRot() );

  rotMat = rotY*rotX;

  // compute ray directions by multiplying each normalized ray direction
  // by the rotation matrix
  int numRays = imageWidth*imageHeight;
  for ( int i=0; i < numRays; i++ )
    rotatedRayDirs[i] = rotMat*rayDirs[i];
}
