
#include <iostream>
#include <string>
#include <stdexcept>
#include <cstdlib>

#include "RayTracerApp.h"
#include "Camera.h"
#include "ListScene.h"
#include "KDTreeScene.h"
#include "TriangleMeshObject.h"
#include "Light.h"
#include "SphereObject.h"

#define DEFAULT_WIDTH        512
#define DEFAULT_HEIGHT       512
#define DEFAULT_BPP          32
#define DEFAULT_PORT         27016 // not used anymore (network support removed)
#define DEFAULT_NUM_THREADS  6

#define RT_WINDOW_TITLE     "Ray-Tracer"


const float RayTracerApp::mouseSensitivity = 10.0f;
const float RayTracerApp::moveSpeed        = 5.0f;


RayTracerApp::RayTracerApp(int argc, char *argv[])  {
  initialized    = false;
  controlEnabled = true;
  rayTracer      = NULL;
  scene          = NULL;
  displaySurface = NULL;

   // initialize options to their default values
  int screenWidth  = DEFAULT_WIDTH;
  int screenHeight = DEFAULT_HEIGHT;
  int numThreads   = DEFAULT_NUM_THREADS;
  bool fullscreen  = false;
  std::string sceneFilename;

  std::cout << "Ray-Tracer By A.J. Tavakoli" << std::endl;

  // parse command line arguments and bail if they're invalid
  if ( !parseArgs(argc, argv, screenWidth, screenHeight,
                  numThreads, fullscreen) ) {
    printHelp(argv[0]);
    return;
  }

  try {
    // initialize display (and SDL)
    initVideo(screenWidth, screenHeight, fullscreen);

    initScene();

    rayTracer = new RayTracer(screenWidth,
                              screenHeight,
                              numThreads,
                              scene);
  } // try
  catch ( RTException &e ) {
    // exit on exceptions
    std::cerr << "error: " << e.what() << std::endl;
    return;
  }
  catch ( std::exception &e ) {
    // exit on exceptions
    std::cerr << "error: " << e.what() << std::endl;
    return;
  }

  initialized = true;
} // RayTracerApp::~RayTracerApp()


RayTracerApp::~RayTracerApp() {
  if ( rayTracer != NULL )
    delete rayTracer;
  if ( scene != NULL )
    delete scene;

  if ( displaySurface )
    SDL_FreeSurface(displaySurface);

  SDL_Quit();

/*
#ifdef WIN32
  // TODO: get rid of this (it's for windows only so I can check debug output in msvc)
  system("pause");
#endif
*/
} // RayTracerApp::~RayTracerApp()


void RayTracerApp::initVideo(int screenWidth,
                             int screenHeight,
                             bool fullscreen) {
  // initialze SDL
  int result = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);

  // was SDL initialization successful?
  if ( -1 == result )
    throw RTException("unable to initialze SDL");

  // initialize video
  if ( fullscreen )
    displaySurface = SDL_SetVideoMode(screenWidth, screenHeight, DEFAULT_BPP,
                                      SDL_HWSURFACE | SDL_FULLSCREEN);
  else
    displaySurface = SDL_SetVideoMode(screenWidth, screenHeight, DEFAULT_BPP,
                                      SDL_HWSURFACE);

  // make sure pointer to display surface is not NULL
  if ( !displaySurface )
    throw RTException("unable to set video mode");

  // set window title
  SDL_WM_SetCaption(RT_WINDOW_TITLE, RT_WINDOW_TITLE);

  // hide mouse cursor
  SDL_ShowCursor(SDL_DISABLE);
}


void RayTracerApp::initScene() {
/*
  // skybox texture names
  std::string skyTextures[6] = {  "scenes/skyboxes/feb_06/skybox_right.tga",
                                  "scenes/skyboxes/feb_06/skybox_left.tga",
                                  "scenes/skyboxes/feb_06/skybox_front.tga",
                                  "scenes/skyboxes/feb_06/skybox_back.tga",
                                  "scenes/skyboxes/feb_06/skybox_top.tga",
                                  "scenes/skyboxes/feb_06/skybox_bottom.tga" };
*/

  // create scene object
  scene = new KDTreeScene;
  //scene = new ListScene;
  
  // create materials
  Material checkerboard;
  checkerboard.ambient = RGBColor<float>(0.8f, 0.8f, 0.8f);
  checkerboard.diffuse = RGBColor<float>(0.8f, 0.8f, 0.8f);
  checkerboard.texture = scene->loadTexture("models/checkerboard_red.tga");
  scene->insertMaterial(checkerboard);
  
  Material shinyGray;
  shinyGray.ambient   = RGBColor<float>(0.5f, 0.5f, 0.5f);
  shinyGray.diffuse   = RGBColor<float>(0.5f, 0.5f, 0.5f);
  shinyGray.specular  = RGBColor<float>(1.0f, 1.0f, 1.0f);
  shinyGray.shininess = 30.0f;
  scene->insertMaterial(shinyGray);
  
  Material redRefl;
  redRefl.ambient      = RGBColor<float>(0.7f, 0.0f, 0.0f);
  redRefl.diffuse      = RGBColor<float>(0.7f, 0.0f, 0.0f);
  redRefl.specular     = RGBColor<float>(1.0f, 1.0f, 1.0f);
  redRefl.shininess    = 30.0f;
  redRefl.reflectivity = 1.0f;
  scene->insertMaterial(redRefl);
  
  Material refrRefl;
  refrRefl.ambient      = RGBColor<float>(0.1f, 0.1f, 0.1f);
  refrRefl.diffuse      = RGBColor<float>(0.1f, 0.1f, 0.1f);
  refrRefl.specular     = RGBColor<float>(1.0f, 1.0f, 1.0f);
  refrRefl.shininess    = 30.0f;
  refrRefl.reflectivity = 0.4f;
  refrRefl.refrIndex    = 1.2f;
  scene->insertMaterial(refrRefl);
  
  // one point light in scene
  Light light;
  light.pos = Point3(-0.5f, 3.0f, -11.0f);
  light.ambient  = RGBColor<float>(0.2f, 0.2f, 0.2f);
  light.diffuse  = RGBColor<float>(0.7f, 0.7f, 0.7f);
  light.specular = RGBColor<float>(1.0f, 1.0f, 1.0f);
  scene->insertLight(light);
  
  // load skybox textures
  //scene->loadSkyBox(skyTextures);
  
  // insert 3d model for floor
  // (remember that one TriangleMesh can be used
  //  for multiple TriangleMeshObjects)
  scene->insertObject( new TriangleMeshObject(*(scene->loadModel("models/floor.msh", 0)),
                                              0,
                                              Point3(0.0f, -2.0f, -10.0f)) );
                                             
  scene->insertObject( new TriangleMeshObject(*(scene->loadModel("models/budda_rot_lowres.msh", 1, 2.0f)),
                                              1,
                                              Point3(-3.0f, -1.27, -14.0)) );

  // insert two spheres (a reflective one and a reflective/refractive one)
  scene->insertObject( new SphereObject(Point3(-3.0f, -1.22f, -11.0f),
                                        0.75f,
                                        2) );
                                        
  scene->insertObject( new SphereObject(Point3(2.0f, -1.22f, -11.0f),
                                        0.75f,
                                        3) );

  scene->build();
}


int RayTracerApp::exec() {
  // if initialization failed (done in constructor), then
  // return failure
  if ( !initialized || !rayTracer )
    return EXIT_FAILURE;

  // these are used in the application's main loop
  bool running = true;
  unsigned int frameCount = 0;
  float avgFPS = 0.0;
  Uint32 lastHandleMouse,
         lastTickCount;
  lastHandleMouse = lastTickCount = SDL_GetTicks();

  // warp mouse to center of window since offset from center of window
  // is used to determine rotation
  // (this avoids an unintentional initial rotation)
  SDL_WarpMouse(rayTracer->getImageWidth()  >> 1,
                rayTracer->getImageHeight() >> 1);

  // main application loop
  while ( running ) {
    // poll for events, and handle each one
    SDL_Event event;
    while ( SDL_PollEvent( &event ) ) {
      switch ( event.type ) {
        case SDL_KEYDOWN:
          if ( SDLK_ESCAPE == event.key.keysym.sym )
            running = false;
          break;
        case SDL_QUIT: running = false; break;
        default: break;
      }
    }

    Uint32 startTicks = SDL_GetTicks();

    if ( controlEnabled ) {
      // process mouse input
      // (at most every 10 ms)
      if ( startTicks > lastHandleMouse + 10 ) {
        handleMouseMotion(float(startTicks - lastHandleMouse)/1000.0f);
        lastHandleMouse = startTicks;
      }

      // process keyboard input
      handleKeyboard(float(startTicks - lastTickCount)/1000.0f);
    }

    // run main function of ray-tracer
    rayTracer->mainLoop();

    // copy image to display
    showImage();
 
    // update FPS average
    float frameTime = float(SDL_GetTicks() - startTicks);
    frameCount++;
    avgFPS += frameTime;

    lastTickCount = startTicks;
  } // while ( running )

  // print out average FPS
  //avgFPS = 1000.0f/( avgFPS/float(frameCount) );
  //std::cout << "average FPS: " << avgFPS << std::endl;

  return EXIT_SUCCESS;
} // RayTracerApp::exec()


void RayTracerApp::handleKeyboard(float dt) {
  Camera &camera = rayTracer->getCamera();
  Matrix3 rotMat;
  rotMat.roty( camera.getYRot() );

  Uint8 *kbd = SDL_GetKeyState(NULL);
  if ( kbd[SDLK_LEFT] == 1 || kbd[SDLK_a] == 1 )
    camera.updatePos( rotMat*Vector3(-moveSpeed*dt, 0.0f, 0.0f) );

  if ( kbd[SDLK_RIGHT] == 1 || kbd[SDLK_d] == 1)
    camera.updatePos( rotMat*Vector3(moveSpeed*dt, 0.0f, 0.0f) );

  if ( kbd[SDLK_UP] == 1 || kbd[SDLK_w] == 1 )
    camera.updatePos( rotMat*Vector3(0.0f, 0.0f, -moveSpeed*dt) );

  if ( kbd[SDLK_DOWN] == 1 || kbd[SDLK_s] == 1 )
    camera.updatePos( rotMat*Vector3(0.0f, 0.0f, moveSpeed*dt) );
    
  if ( kbd[SDLK_PAGEUP] == 1 )
    camera.updatePos( Vector3(0.0f, moveSpeed*dt, 0.0f) );
  
  if ( kbd[SDLK_PAGEDOWN] == 1 )
    camera.updatePos( Vector3(0.0f, -moveSpeed*dt, 0.0f) );
} // RayTracerApp::handleKeyboard()


void RayTracerApp::handleMouseMotion(float dt) {
  // to store x, y coordinates of mouse cursor
  int x = 0,
      y = 0;

  // get mouse button state
  SDL_GetMouseState(&x, &y);

  ///// handle mouselook /////

  Camera &camera = rayTracer->getCamera();

  // angle differences
  float dTheta = 0.0f,
        dPhi   = 0.0f;

  int imageWidth  = rayTracer->getImageWidth(),
      imageHeight = rayTracer->getImageHeight();

  int halfWidth  = imageWidth  >> 1,
      halfHeight = imageHeight >> 1;

  // horizontal and vertical differences in cursor position
  float dx = float(x - halfWidth),
        dy = float(y - halfHeight);

  // if cursor has moved horizontally
  if ( dx )
    dTheta = dx/float(halfWidth);

  // if cursor has moved vertically
  if ( dy )
    dPhi = dy/float(halfHeight);

  if ( dTheta != 0.0 || dPhi != 0.0 ) {
    // update camera orientation
    camera.updateXRot(-dPhi*mouseSensitivity*dt);
    camera.updateYRot(-dTheta*mouseSensitivity*dt);

    // restrict camera orientation
    if ( camera.getXRot() > 1.57079633f )
      camera.setXRot(1.57079633f);
    else if ( camera.getXRot() < -1.57079633f )
      camera.setXRot(-1.57079633f);
  }
  
  if ( dx || dy )
    rayTracer->updateRayDirs();

  SDL_WarpMouse(halfWidth, halfHeight);
} // RayTracerApp::handleMouseMotion()


// parse command line arguments
bool RayTracerApp::parseArgs(int argc,
                             char *argv[],
                             int &screenWidth,
                             int &screenHeight,
                             int &numThreads,
                             bool &fullscreen) {
  // parse each command line argument
  for ( int i=1; i < argc; i++) {
    if ( strcmp(argv[i], "--r") == 0 ) {
      // sets resolution
      // make sure don't go past end of argv while looking for width, height, and bpp
      // TODO: make sure arguments are integers
      if ( argc <= i + 2 ||
           strncmp(argv[i+1], "--", 2) == 0 ||
           strncmp(argv[i+2], "--", 2) == 0 ) {
        std::cerr << "invalid arguments: expected --r <width> <height>"
                  << std::endl;
        return false;
      }

      // convert arguments to integers
      screenWidth  = atoi(argv[++i]);
      screenHeight = atoi(argv[++i]);
    }
    else if ( strcmp(argv[i], "--t") == 0 ) {
      // specifies number of threads
      if ( argc <= i + 1 || strncmp(argv[i+1], "--", 2) == 0 ) {
        std::cerr << "invalid arguments: expected --t <# of threads>"
                  << std::endl;
        return false;
      }

      // convert argument to integer
      numThreads = atoi(argv[++i]);
    }
    else if ( strcmp(argv[i], "--f") == 0 ) {
      fullscreen = true;
    }
    else {
      // unknown argument
      std::cerr << "invalid argument: " << argv[i] << std::endl;
      return false;
    }
  } // for ( command-line arguments )

  // make sure arguments are valid

  // width, height must be greater than 0
  if ( screenWidth < 1 || screenHeight < 1 ) {
    std::cerr << "invalid arguments: screen dimension must be greater than 0"
              << std::endl;
    return false;
  }

  // # of threads must be greater than 0
  if ( numThreads < 1 ) {
    std::cerr << "invalid arguments: # of threads must be at least 1"
              << std::endl;
    return false;
  }

  // indicate arguments were valid
  return true;
} // RayTracerApp::parseArgs()
  
  
// prints out command line usage information
void RayTracerApp::printHelp(const char *cmd) {
  std::cout << "usage: " << cmd << " <options>" << std::endl
            << "<options> are:" << std::endl
            << "--r <width> <height>        set image resolution to <width>x<height>" << std::endl
            << "--f                         fullscreen mode" << std::endl
            << "--t <# of threads>          set number of threads to <# of threads>" << std::endl
            << std::endl;
} // RayTracerApp::printHelp()


void RayTracerApp::showImage() {
  // sanity check
  if ( !rayTracer || !displaySurface )
    return;

  const Uint8 *imageData = rayTracer->getImage();

  // make sure image pointer is non-NULL
  if ( !imageData )
    return;

  int imageWidth  = rayTracer->getImageWidth(),
      imageHeight = rayTracer->getImageHeight();

  if ( SDL_MUSTLOCK(displaySurface) &&
       SDL_LockSurface(displaySurface) == -1 )
    throw std::runtime_error("could not lock screen surface");

  Uint8 *buffer = (Uint8 *)displaySurface->pixels;

  // finally, copy the image
  for ( int y=0; y < imageHeight; y++ ) {
    int srcBase = imageWidth*y;
    int dstBase = (imageHeight-y-1)*displaySurface->pitch;
    for ( int x=0; x < imageWidth; x++ ) {
      int srcIndex = (srcBase + x)*3;
      int dstIndex = int(dstBase + x*displaySurface->format->BytesPerPixel);

      buffer[dstIndex  ] = imageData[srcIndex+2];
      buffer[dstIndex+1] = imageData[srcIndex+1];
      buffer[dstIndex+2] = imageData[srcIndex  ];
    }
  }

  // if display surface must be locked, unlock it
  if ( SDL_MUSTLOCK(displaySurface) )
    SDL_UnlockSurface(displaySurface);

  SDL_UpdateRect(displaySurface, 0, 0, 0, 0);
} // RayTracerApp::showImage()
