
#include "Scene.h"
#include "SphereObject.h"
#include "TriangleMeshKDTree.h"
#include "TriangleMeshObject.h"
#include <map>


Scene::Scene():
     skyBox(NULL) {

}


Scene::~Scene() {
  clearScene();
}


void Scene::clearScene() {
  // delete all textures
  for ( size_t i=0; i < textures.size(); i++ )
    if ( textures[i] )
      delete textures[i];
      
  // delete all meshes
  for ( size_t i=0; i < meshes.size(); i++ )
    if ( meshes[i] )
      delete meshes[i];
      
  // delete skybox
  if ( skyBox )
    delete skyBox;
    
  textures.clear();
  meshes.clear();
  skyBox = NULL;
}


// This method is called by the implementation of the lightContrib() method.
// It calculates the contribution from a single light.  This method is a
// member of the Scene base class since its behavior should not be overridden;
// it is the same for all types of acceleration structures.
RGBColor<float> Scene::singleLightContrib(const Light &light,
                                          const Point3 &p,
                                          const Ray &ray,
                                          const Vector3 &n,
                                          const Material &material) {
  // this will store the light's contribution
  RGBColor<float> contrib;

  // add ambient contribution
  contrib.c[0] = light.ambient.c[0]*material.ambient.c[0];
  contrib.c[1] = light.ambient.c[1]*material.ambient.c[1];
  contrib.c[2] = light.ambient.c[2]*material.ambient.c[2];

  // compute vector from hitpoint to pointlight (light vector)
  Vector3 toLight = light.pos - p;

  // compute distance from hitpoint to pointlight
  float dist = toLight.mag();

  // if distance from hitpoint to pointlight is close to zero
  // then return only the ambient contribution
  if ( dist < RT_EPSILON )
    return contrib;

  // normalize light vector
  toLight *= 1.0f/dist;

  // create a ray starting at hitpoint and with its direction
  // pointing toward the light
  Ray lightRay;
  lightRay.setOrigin(p + toLight*RT_EPSILON);
  lightRay.setDirDontNormalize(toLight); // don't normalize ray direction
                                         // since it already has been normalized

  // is the light ray occluded?
  if ( occlusionTest(lightRay, dist) )
    return contrib; // light is occluded by an object
                    // so return only ambient contribution

  // light ray is not occluded, so add a contribtution from this light

  // diffuse contribution
  float nDotL = n*toLight;
  if ( nDotL > 0 ) {
    contrib.c[0] += light.diffuse.c[0]*material.diffuse.c[0]*nDotL;
    contrib.c[1] += light.diffuse.c[1]*material.diffuse.c[1]*nDotL;
    contrib.c[2] += light.diffuse.c[2]*material.diffuse.c[2]*nDotL;
  }
  
  // specular contribution
  Vector3 r = toLight - n*(2.0f*nDotL);
  float vDotR = ray.getDir()*r;
  if ( vDotR > 0.0f ) {
    float specPow = powf(vDotR, material.shininess);
    contrib.c[0] += specPow*light.specular.c[0]*material.specular.c[0];
    contrib.c[1] += specPow*light.specular.c[1]*material.specular.c[1];
    contrib.c[2] += specPow*light.specular.c[2]*material.specular.c[2];
  }

  // compute attenuation
  float att = 1.0f;
  if ( dist > 0.0f ) {
    att = 1.0f/(light.constAtt +
                light.linearAtt/dist +
                light.quadAtt/(dist*dist));
  }
  // return contribution from this light
  return contrib*att;
}


Texture2D *Scene::loadTexture(const char *filename) {
  // sanity check
  if ( !filename )
    return NULL;

  Texture2D *texture = NULL;

  TGAFile tga;
  if ( tga.load(filename) ) {
    texture = new Texture2D(tga);
    textures.push_back(texture);
  }
  else {
    std::string errorMsg("Scene::loadTexture(): unable to load ");
    errorMsg += filename;
    throw RTException(errorMsg);
  }

  return texture;
}


// TODO: get rid of matID
TriangleMesh *Scene::loadModel(const char *filename, int matID, float scale) {
  // sanity check
  if ( !filename )
    return NULL;
    
  TriangleMesh *mesh = new TriangleMeshKDTree(matID, filename, scale);
  
  meshes.push_back(mesh);

  return mesh;
}


void Scene::loadSkyBox(const std::string textureNames[6]) {
  Texture2D *skyBoxTextures[6] = { NULL, NULL, NULL, NULL, NULL, NULL };
  
  // load each skybox texture
  for ( int i=0; i < 6; i++ ) {
    if ( !textureNames[i].empty() ) {
      TGAFile tga;
      if ( tga.load(textureNames[i].c_str()) == false ) {
        std::string errorMsg = "Scene::loadSkyBox(): unable to load skybox texture: ";
        errorMsg += textureNames[i];
        throw RTException(errorMsg);
      }
      
      Texture2D *skyBoxTexture = new Texture2D(tga);
      
      if ( !skyBoxTexture )
        throw RTException("Scene::loadSkyBox(): unable to allocate storage for skybox texture");
        
      // add texture to array of textures (so it will be deleted when 
      // scene is cleared)
      textures.push_back(skyBoxTexture);
      
      skyBoxTextures[i] = skyBoxTexture;
    } // if (texture filename is not empty string)
  }
  
  // get rid of the current skybox if there already is one
  if ( skyBox )
    delete skyBox;
   
  // create new skybox
  skyBox = new SkyBox(skyBoxTextures);
  
  if ( !skyBox )
    throw RTException("Scene::loadSkyBox(): unable to allocate storage for skybox");
}


void Scene::insertMaterial(const Material &material) {
  materials.push_back(material);
}
