
#include <cstdlib>
#include <cmath>
#include <stdexcept>
#include <iostream>
#include "Math3D/numeric.h"
#include "Math3D/Vector3.h"
#include "Texture2D.h"


Texture2D::Texture2D():
     width(0),
     height(0),
     fWidth(0.0f),
     fHeight(0.0f),
     image(NULL) {

} // Texture2D::Texture2D()


// constructs texture from a TGA image
Texture2D::Texture2D(const TGAFile &tga) {
  width = height = 0;
  image = NULL;

  // make sure TGA image contains data
  if ( tga.getData() == NULL || tga.getWidth() <= 0 || tga.getHeight() <= 0 )
    return;

  // tga must be either 24 or 32-bit
  if ( tga.getBPP() != 24 && tga.getBPP() != 32 )
    throw std::runtime_error("Texture2D::Texture2D(): color depth of "
                             "image not supported");

  // allocate storage for image
  image = new RGBColor<float>[tga.getWidth()*tga.getHeight()];

  // was memory allocated successfully?
  if ( NULL == image )
    throw std::runtime_error("Texture2D::Texture2D(): unable to allocate "
                             "memory for texture");

  // store dimensions of image
  width   = tga.getWidth();
  height  = tga.getHeight();
  fWidth  = float(width-1);
  fHeight = float(height-1);

  // copy image data
  int tgaStep = tga.getBPP() >> 3;
  const unsigned char *tgaData = tga.getData();
  for ( int y=0; y < height; y++ ) {
    int base = y*width;
    int tgaBase = base*tgaStep;

    for ( int x=0; x < width; x++ ) {
      int tgaIndex = tgaBase + x*tgaStep,
          index    = base + x;
      image[index].c[0] = float(tgaData[tgaIndex])/255.0f;
      image[index].c[1] = float(tgaData[tgaIndex+1])/255.0f;
      image[index].c[2] = float(tgaData[tgaIndex+2])/255.0f;
    } // for x
  } // for y
} // Texture2D::Texture2D()


Texture2D::~Texture2D() {
  if ( image )
    delete [] image;
} // Texture2D::Texture2D()


RGBColor<float> Texture2D::sample(float u, float v) const {
  RGBColor<float> color(0.0f, 0.0f, 0.0f);

  // make sure texture coords are in range and
  // implement texture repeating
  // not supporting clamping since it wont be used in this program
  u = fabs(u);
  v = fabs(v);

  double intPart; // used only because modf() requires this argument

  if ( u > 1.0f )
    u = (float)modf(u, &intPart);
  if ( v > 1.0f )
    v = (float)modf(v, &intPart);

  // convert sample coordinates to integers
  // TODO: maybe provide some options for interpolation
  int iU = int(u*fWidth),
      iV = int(v*fHeight);

  if ( iU < width && iV < height ) {
    unsigned int index = iV*width + iU;
    color.c[0] = (float)image[index].c[0];
    color.c[1] = (float)image[index].c[1];
    color.c[2] = (float)image[index].c[2];
  }

  return color;
} // Texture2D::sample()


RGBColor<float> Texture2D::sampleBilinear(float u, float v) const {
  // for the sake of speed, do not check to see if there is data

  // check if sample coordinates are not in range [u, v]
  // (implements texture repeating)
  // TODO: add an option for this (support clamping also)
  if ( u > 1.0f )
    u = u - floorf(u);
  if ( v > 1.0f )
    v = v - floorf(v);
  if ( u < 0.0f ) {
    u = -u;
    u = (u - floorf(u));
  }
  if ( v < 0.0f ) {
    v = -v;
    v = (v - floorf(v));
  }

  ///// do bilinear interpolation /////

  // compute sample coordinates in pixels
  float pxlU = u*fWidth,
        pxlV = v*fHeight;

  // compute floor and ceiling of sample coordinates for interpolating
  float floorU = floorf(pxlU),
        floorV = floorf(pxlV),
        ceilU  = ceilf(pxlU),
        ceilV  = ceilf(pxlV);

  // convert sample coordinates to integers
  int iFloorU = int(floorU),
      iFloorV = int(floorV),
      iCeilU  = int(ceilU),
      iCeilV  = int(ceilV);

  // make sure ceiling is in bounds. dimensions must be at least 1!
  // for the sake of speed there is no check, although the impact
  // would likely be negligible
  if ( iCeilU >= width )
    iCeilU = width-1;
  if ( iCeilV >= height )
    iCeilV = height-1;

  int topLeftIndex     = iFloorV*width + iFloorU,
      bottomLeftIndex  = iCeilV*width  + iFloorU,
      topRightIndex    = iFloorV*width + iCeilU,
      bottomRightIndex = iCeilV*width  + iCeilU;

  // take samples
  RGBColor<float> topLeftSample( (float)image[topLeftIndex].c[0],
                                 (float)image[topLeftIndex].c[1],
                                 (float)image[topLeftIndex].c[2] ),
                  bottomLeftSample( (float)image[bottomLeftIndex].c[0],
                                    (float)image[bottomLeftIndex].c[1],
                                    (float)image[bottomLeftIndex].c[2] ),
                  topRightSample( (float)image[topRightIndex].c[0],
                                  (float)image[topRightIndex].c[1],
                                  (float)image[topRightIndex].c[2] ),
                  bottomRightSample( (float)image[bottomRightIndex].c[0],
                                     (float)image[bottomRightIndex].c[1],
                                     (float)image[bottomRightIndex].c[2]);

  // get fractional parts of texture coordinates
  pxlU -= floorU;
  pxlV -= floorV;

  return topLeftSample     * (  (1.0f-pxlU) * (1.0f-pxlV) ) +
         topRightSample    * (  pxlU        * (1.0f-pxlV) ) +
         bottomLeftSample  * (  (1.0f-pxlU) * pxlV        ) +
         bottomRightSample * (  pxlU        * pxlV        );
} // Texture2D::sampleBilinear()
