
#include <cstdio>
#include <climits>
#include "Math3D/intersect.h"
#include "TriangleMesh.h"


TriangleMesh::TriangleMesh(int _matID, const char *filename, float scale):
     matID(_matID),
     aabb( Point3(0.0f, 0.0f, 0.0f), Point3(0.0f, 0.0f, 0.0f) ),
     numTris(0),
     numVertices(0),
     vertIndices(NULL),
     surfaceTangents(NULL),
     surfaceBitangents(NULL),
     surfaceNormals(NULL),
     triAxis(NULL),
     triAttrs(NULL),
     vertPositions(NULL),
     vertTexCoords(NULL),
     vertNormals(NULL),
     vertBitangents(NULL),
     vertTangents(NULL) {
  load(filename, scale);
}


TriangleMesh::~TriangleMesh() {
  clear();
}


void TriangleMesh::clear() {
  if ( vertIndices ) {
    delete [] vertIndices;
    vertIndices = NULL;
  }

  if ( vertPositions ) {
    delete [] vertPositions;
    vertPositions = NULL;
  }

  if ( vertTexCoords ) {
    delete [] vertTexCoords;
    vertTexCoords = NULL;
  }

  if ( vertNormals ) {
    delete [] vertNormals;
    vertNormals = NULL;
  }
  
  if ( vertBitangents ) {
    delete [] vertBitangents;
    vertBitangents = NULL;
  }
  
  if ( vertTangents ) {
    delete [] vertTangents;
    vertTangents = NULL;
  }

  if ( surfaceTangents ) {
    delete [] surfaceTangents;
    surfaceTangents = NULL;
  }

  if ( surfaceBitangents ) {
    delete [] surfaceBitangents;
    surfaceBitangents = NULL;
  }

  if ( surfaceNormals ) {
    delete [] surfaceNormals;
    surfaceNormals = NULL;
  }

  if ( triAxis ) {
    delete [] triAxis;
    triAxis = NULL;
  }

  if ( triAttrs ) {
    delete [] triAttrs;
    triAttrs = NULL;
  }

  numTris = 0;
  numVertices = 0;
}
  

void TriangleMesh::processTriangle(int i) {
  // make sure there is mesh data before proceeding
  if ( i > numTris || !vertIndices || !vertPositions || !surfaceNormals )
    return;

  int v0Index = i*3;
  Point3 v0 = vertPositions[ vertIndices[v0Index  ] ],
         v1 = vertPositions[ vertIndices[v0Index+1] ],
         v2 = vertPositions[ vertIndices[v0Index+2] ];

  // calculate normal vector
  // (assuming right-handed coordinate system and counter-clockwise
  //  vertex order)
  Vector3 e0 = v1 - v0,
          e1 = v2 - v0;
  Vector3 normal = Vector3::cross(e0, e1);
  normal.normalize();
  surfaceNormals[i] = normal;

  // determine dominant axis
  unsigned char axis = Math3D::rabs(normal[0]) > Math3D::rabs(normal[1]) ? 0 : 1;
  if ( Math3D::rabs(normal[2]) > Math3D::rabs(normal[axis]) )
    axis = 2;

  unsigned char uAxis = (axis+1)%3,
                vAxis = (axis+2)%3;
  axis |= uAxis << 2;
  axis |= vAxis << 4;

  triAxis[i] = axis;

  // compute attributes that are used for ray-triangle test
  unsigned char k = axis & 0x03;

  Tri &triAttr = triAttrs[i];

  // first line equation
  float reci = 1.0f/(e1[uAxis]*e0[vAxis] - e1[vAxis]*e0[uAxis]);
  triAttr.bnu  = e1[uAxis]*reci;
  triAttr.bnv  = -e1[vAxis]*reci;

  // second line equation
  triAttr.cnu = e0[vAxis]*reci;
  triAttr.cnv = -e0[uAxis]*reci;

  float krec = 1.0f/normal[k];
  triAttr.nu = normal[uAxis]*krec;
  triAttr.nv = normal[vAxis]*krec;
  triAttr.nd = (normal*v0)*krec;
}


// computes vertex normals, must be called after surface normals have
// been computed
void TriangleMesh::computeVertexNormals(int numVerts) {
  // make sure there is data
  if ( numTris <= 0 || numVerts <= 0 || !surfaceNormals )
    return;

  // create temporary storage for an array that will contain how many
  // faces include each vertex
  std::vector<int> contrib(numVerts);

  // zero out temporary storage and vertex normals
  for ( int i=0; i < numVerts; i++ ) {
    contrib[i] = 0;
    vertNormals[i] = Vector3(0.0f, 0.0f, 0.0f);
  }

  // iterate through faces and find out how many faces contain each vertex
  for ( int i=0; i < numTris; i++ ) {
    int v0Index = i*3;
    const Vector3 &surfaceNormal = surfaceNormals[i];
    for ( int j=0; j < 3; j++ ) {
      int vertIndex = vertIndices[v0Index+j];
      contrib[vertIndex]++;
      vertNormals[vertIndex] += surfaceNormal;
    }
  }

  for ( int i=0; i < numVerts; i++ ) {
    int      n      = contrib[i];
    Vector3 &normal = vertNormals[i];
    if ( n > 0 ) // this should always be true... but check anyway
      normal /= float(n);
    if ( normal.sqrMag() > 0.0001f )
      normal.normalize();
  }
}


void TriangleMesh::buildAABB() {
  // make sure there is data before proceeding
  if ( numTris <= 0 || !vertIndices || !vertPositions )
     return;

  Point3 minP( RT_FLT_MAX,  RT_FLT_MAX,  RT_FLT_MAX ),
         maxP(-RT_FLT_MAX, -RT_FLT_MAX, -RT_FLT_MAX );

  int numVerts = numTris*3;
  for ( int i=0; i < numVerts; i++ ) {
    const Point3 &v = vertPositions[ vertIndices[i] ];
    for ( int j=0; j < 3; j++ ) {
      if ( v[j] < minP[j] )
        minP[j] = v[j];
      if ( v[j] > maxP[j] )
        maxP[j] = v[j];
    }
  }
  
  // make sure box has some thickness along each axis
  for ( int i=0; i < 3; i++ )
    if ( maxP[i] - minP[i] < RT_EPSILON )
      maxP[i] = minP[i] + RT_EPSILON;

  aabb = AABB(minP, maxP);
}


void TriangleMesh::allocateStorage(int newNumTris,
                                   int newNumVerts,
                                   bool allocTC,
                                   bool allocVertNormals) {
  // clear data if it hasn't been already
  clear();

  // allocate storage for faces
  vertIndices = new int[newNumTris*3];

  if ( !vertIndices )
    throw RTException("TriangleMesh::allocateStorage(): unable to allocate storage for faces");

  // allocate storage for vertices
  vertPositions = new Point3[newNumVerts];

  if ( !vertPositions ) {
    clear();
    throw RTException("TriangleMesh::allocateStorage(): unable to allocate storage for vertices");
  }

  // allocate storage for surface normals
  surfaceNormals = new Vector3[newNumTris];

  if ( !surfaceNormals ) {
    clear();
    throw RTException("TriangleMesh::allocateStorage(): unable to allocate storage for surface normals");
  }

  // allocate storage for axis
  triAxis = new unsigned char[newNumTris];

  if ( !triAxis ) {
    clear();
    throw RTException("TriangleMesh::allocateStorage(): unable to allocate storage for triAxis");
  }

  // allocate storage for triangle attributes
  triAttrs = new Tri[newNumTris];

  if ( !triAttrs ) {
    clear();
    throw RTException("TriangleMesh::allocateStorage(): unable to allocate "
                      "storage for triangle attributes");
  }

  // zero everything out
  memset(vertIndices, 0, sizeof(int)*newNumTris*3);
  memset(vertPositions, 0, sizeof(Point3)*newNumVerts);
  memset(surfaceNormals, 0, sizeof(Vector3)*newNumTris);
  memset(triAxis, 0, sizeof(unsigned char)*newNumTris);
  memset(triAttrs, 0, sizeof(Tri)*newNumTris);

  if ( allocTC ) {
    vertTexCoords = new Point2[newNumVerts];

    if ( !vertTexCoords ) {
      clear();
      throw RTException("TriangleMesh::allocateStorage(): unable to allocate storage for texture coordinates");
    }

    memset(vertTexCoords, 0, sizeof(Point2)*newNumVerts);
  }

  if ( allocVertNormals ) {
    // allocate storage for vertex normals
    vertNormals = new Vector3[newNumVerts];

    if ( !vertNormals ) {
      clear();
      throw RTException("TriangleMesh::allocateStorage(): unable to allocate storage for vertex normals");
    }

    memset(vertNormals, 0, sizeof(Vector3)*newNumVerts);
  }

  numTris     = newNumTris;
  numVertices = newNumVerts;
}


// this function uses extension of filename to determine which format
// to load
void TriangleMesh::load(const char *filename, float scale) {
  // make sure filename is not NULL
  if ( NULL == filename )
    throw RTException("TriangleMesh::load(NULL): filename is NULL");

  // find last . in filename
  // I avoid strtok() because it's not thread-safe (although it probably
  // does not matter in this case)
  size_t len = strlen(filename);
  const char *lastDot = NULL;
  for ( size_t i=0; i < len; i++ )
    if ( filename[i] == '.' )
      lastDot = &filename[i];

  // copy file extension into temporary buffer and then convert to lower case
  char fileExt[1024] = "";
  strcpy(fileExt, lastDot);
  len = strlen(fileExt);
  for ( size_t i=0; i < len; i++ )
    fileExt[i] = tolower(fileExt[i]);

  // determine which file type to load based on extension
  if ( strcmp(fileExt, ".obj") == 0 ) {
    // no longer supported
    throw RTException("TriangleMesh::load(): obj models no longer supported");
    //loadOBJ(filename, scale);
  }
  else {
    // fall back to loading a MSH file
    loadMSH(filename, scale);
  }
}


void TriangleMesh::loadMSH(const char *filename, float scale) {
  // make sure filename is not NULL  
  if ( NULL == filename )
    throw RTException("TriangleMesh::loadMSH(): filename is NULL");

  // this string will go at the beginning of all of the error messages
  // thrown in this method
  std::string errorMsgStart("TriangleMesh::loadMSH(");
  errorMsgStart += filename;
  errorMsgStart += "): ";

  // this should not be called if there is already data in mesh
  if ( numTris > 0 || vertIndices || vertPositions ) {
    std::string errorMsg(errorMsgStart);
    errorMsg += "mesh data has already been loaded";
    throw RTException(errorMsg);
  }

  // try to open file for reading
  FILE *fin = fopen(filename, "r");

  // was file opened successfully?
  if ( NULL == fin ) {
    std::string errorMsg(errorMsgStart);
    errorMsg += "unable to open file for reading";
    throw RTException(errorMsg);
  }

  // read in number of triangles
  int n = 0;
  if ( fscanf(fin, "TRIANGLE_MESH NUM_TRIANGLES %d\n", &n) != 1 ) {
    fclose(fin);
    std::string errorMsg(errorMsgStart);
    errorMsg += "unable to read in number of triangles in mesh";
    throw RTException(errorMsg);
  }

  // make sure number of triangles is greater than zero
  if ( n < 0 ) {
    fclose(fin);
    std::string errorMsg(errorMsgStart);
    errorMsg += "number of triangles in mesh is invalid";
    throw RTException(errorMsg);
  }

  int newNumTris = n;

  // read in number of vertices
  n = 0;
  if ( fscanf(fin, "NUM_VERTICES %d\n", &n) != 1 ) {
    fclose(fin);
    clear();
    std::string errorMsg(errorMsgStart);
    errorMsg += "unable to read in number of vertices";
    throw RTException(errorMsg);
  }

  if ( n < 0 ) {
    fclose(fin);
    clear();
    std::string errorMsg(errorMsgStart);
    errorMsg += "number of vertices is negative";
    throw RTException(errorMsg);
  }

  int newNumVerts = n;

  // read in HAS_TEXCOORDS value
  int hasTexCoords = 0;
  if ( fscanf(fin, "HAS_TEXCOORDS %d", &hasTexCoords) != 1 ) {
    fclose(fin);
    clear();
    std::string errorMsg(errorMsgStart);
    errorMsg += "expected HAS_TEXCOORDS followed by a value";
    throw RTException(errorMsgStart);
  }

  // make sure value of HAS_TEXCOORDS is correct
  if ( hasTexCoords != 0 && hasTexCoords != 1 ) {
    fclose(fin);
    clear();
    std::string errorMsg(errorMsgStart);
    errorMsg += "HAS_TEXCOORDS value is invalid (must be 0 or 1)";
    throw RTException(errorMsg);
  }

  allocateStorage(newNumTris, newNumVerts, hasTexCoords == 1, true);

  // read in triangles
  for ( int i=0; i < numTris*3; i+=3 ) {
    if ( fscanf(fin, "%d %d %d", &vertIndices[i], &vertIndices[i+1], &vertIndices[i+2]) != 3 ) {
      fclose(fin);
      clear();
      char errorMsg[1024];
      sprintf(errorMsg, "%sunable to read in triangle #%d", errorMsgStart, i/3);
      throw RTException(errorMsg);
    }
  }

  // read in vertices
  for ( int i=0; i < newNumVerts; i++ ) {
    Point3 &v = vertPositions[i];
    if ( fscanf(fin, "%f %f %f", &v[0], &v[1], &v[2]) != 3 ) {
      fclose(fin);
      clear();
      char errorMsg[1024];
      sprintf(errorMsg, "%sunable to read in vertex #%d", errorMsgStart, i);
      throw RTException(errorMsg);
    }
  }

  // scale vertices if scale argument is not 1
  if ( scale < 0.99f || scale > 1.01f ) {
    for ( int i=0; i < newNumVerts; i++ ) {
      Point3 &v = vertPositions[i];
      v[0] *= scale;
      v[1] *= scale;
      v[2] *= scale;
    }
  }

  if ( hasTexCoords ) {
    // read in texture coords
    for ( int i=0; i < newNumVerts; i++ ) {
      Point2 &tc = vertTexCoords[i];
      if ( fscanf(fin, "%f %f", &tc[0], &tc[1]) != 2 ) {
        fclose(fin);
        clear();
        char errorMsg[1024];
        sprintf(errorMsg, "%sunable to read in texture coordinate #%d", errorMsgStart, i);
        throw RTException(errorMsg);
      }
    }
  } // if (has texture coordinates)

  // close input file
  fclose(fin);

  // calculate information for each triangle
  for ( int i=0; i < numTris; i++ )
    processTriangle(i);

  // compute vertex normals
  computeVertexNormals(newNumVerts);

  buildAABB();
}


bool intersectRayAABB(const Ray &ray, const AABB &aabb,
                      float *tNearOut, float *tFarOut) {
  Point3  origin = ray.getOrigin(),
          minP   = aabb.getMin(),
          maxP   = aabb.getMax();
  Vector3 dir    = ray.getDir();

  float t1    = 0.0,
        t2    = 0.0,
        tNear = -RT_FLT_MAX,
        tFar  = RT_FLT_MAX;

  // test against planes in x-axis
  // is ray direction parallel to x-axis?
  if ( dir.m[0] <= 0.0001f && dir.m[0] >= -0.0001f ) {
    // ray direction is parallel to x-axis so the ray can only
    // be intersecting the box if the origin is between the two
    // 'slabs' (planes)
    if ( origin.m[0] < minP.m[0] || origin.m[0] > maxP.m[0] )
      return false;
  }
  else {
    // compute intersection distance of planes
    t1 = (minP.m[0] - origin.m[0])/dir.m[0];
    t2 = (maxP.m[0] - origin.m[0])/dir.m[0];

    // t1 should be nearest intersection
    if ( t1 > t2 ) {
      //swap(t1, t2);
      float temp = t1;
      t1 = t2;
      t2 = temp;
    }
    // we want largest tNear
    if ( t1 > tNear )
      tNear = t1;
    // we want smallest tFar
    if ( t2 < tFar )
      tFar = t2;
    // if box is missed or if box is behind ray return false
    if ( tNear > tFar || tFar < 0.0f )
      return false;
  } // else

  // test y planes
  // first check to see if we are parallel to x planes
  if ( dir.m[1] <= 0.0001f && dir.m[1] >= -0.0001f ) {
    // are we at least between slabs? if not then we bail
    if ( origin.m[1] < minP.m[1] || origin.m[1] > maxP.m[1] )
      return false;
  }
  else  {
    // compute intersection distance of planes
    t1 = (minP.m[1] - origin.m[1])/dir.m[1];
    t2 = (maxP.m[1] - origin.m[1])/dir.m[1];

    // t1 should be nearest intersection
    if ( t1 > t2 ) {
      //swap(t1, t2);
      float temp = t1;
      t1 = t2;
      t2 = temp;
    }
    // we want largest tNear
    if ( t1 > tNear )
      tNear = t1;
    // we want smallest tFar
    if ( t2 < tFar )
      tFar = t2;
    // if box is missed or if box is behind ray return false
    if ( tNear > tFar || tFar < 0.0f )
      return false;
  } // else

  // test z planes
  // first check to see if we are parallel to x planes
  if ( dir.m[2] <= 0.0001f && dir.m[2] >= -0.0001f ) {
    // are we at least between slabs? if not then we bail
    if ( origin.m[2] < minP.m[2] || origin.m[2] > maxP.m[2] )
      return false;
  }
  else {
    // compute intersection distance of planes
    t1 = (minP.m[2] - origin.m[2])/dir.m[2];
    t2 = (maxP.m[2] - origin.m[2])/dir.m[2];

    // t1 should be nearest intersection
    if (t1 > t2) {
      //swap(t1, t2);
      float temp = t1;
      t1 = t2;
      t2 = temp;
    }
    // we want largest tNear
    if (t1 > tNear)
      tNear = t1;
    // we want smallest tFar
    if (t2 < tFar)
      tFar = t2;
    // if box is missed or if box is behind ray return false
    if ( tNear > tFar || tFar < 0.0f )
      return false;
  } // else

  // get entry and exit points into cube
  if ( tNearOut ) *tNearOut = tNear;
  if ( tFarOut  ) *tFarOut  = tFar;

  // ray intersects AABB
  return true;
}


void TriangleMesh::computeBaryCentricCoords(int triIndex,
                                            const Point3 &p,
                                            float b[]) const {
  // get dominant axis of normal
  unsigned char axis = triAxis[triIndex] & 0x03;

  // get vertices of triangle
  int v0Index = triIndex*3;
  Point3 vert0 = vertPositions[ vertIndices[v0Index  ] ],
         vert1 = vertPositions[ vertIndices[v0Index+1] ],
         vert2 = vertPositions[ vertIndices[v0Index+2] ];

  float u1, u2, u3, u4,
        v1, v2, v3, v4;

  if ( 0 == axis ) {
    // x-axis is dominant, project onto yz
    u1 = vert0.m[1] - vert2.m[1];
    u2 = vert1.m[1] - vert2.m[1];
    u3 = p.m[1] - vert0.m[1];
    u4 = p.m[1] - vert2.m[1];

    v1 = vert0.m[2] - vert2.m[2];
    v2 = vert1.m[2] - vert2.m[2];
    v3 = p.m[2] - vert0.m[2];
    v4 = p.m[2] - vert2.m[2];
  }
  else if ( 1 == axis ) {
    // y-axis is dominant, project onto xz
    u1 = vert0.m[2] - vert2.m[2];
    u2 = vert1.m[2] - vert2.m[2];
    u3 = p.m[2] - vert0.m[2];
    u4 = p.m[2] - vert2.m[2];

    v1 = vert0.m[0] - vert2.m[0];
    v2 = vert1.m[0] - vert2.m[0];
    v3 = p.m[0] - vert0.m[0];
    v4 = p.m[0] - vert2.m[0];
  }
  else if ( 2 == axis ) {
    // z-axis is dominant, project onto xy
    u1 = vert0.m[0] - vert2.m[0];
    u2 = vert1.m[0] - vert2.m[0];
    u3 = p.m[0] - vert0.m[0];
    u4 = p.m[0] - vert2.m[0];

    v1 = vert0.m[1] - vert2.m[1];
    v2 = vert1.m[1] - vert2.m[1];
    v3 = p.m[1] - vert0.m[1];
    v4 = p.m[1] - vert2.m[1];
  }

  // TODO: check for divide by zero?
  float invDenom = 1.0f/(v1*u2 - v2*u1);

  b[0] = (v4*u2 - v2*u4)*invDenom;
  b[1] = (v1*u3 - v3*u1)*invDenom;
  b[2] = 1.0f - b[0] - b[1];
}


// returns the bounding box of a single triangle in mesh
AABB TriangleMesh::calcTriAABB(int i) {
  // make sure index of triangle is in bounds
  if ( i >= numTris )
    throw RTException("TriangleMesh::calcTriAABB(): index is greater than number of triangles");

  // initially set min and max of box to first vertex
  int v0Index = i*3;
  const Point3 &v0 = vertPositions[ vertIndices[v0Index] ];
  Point3 minP = v0,
         maxP = v0; 

  // see if any component of 2nd or 3rd vertex extend further
  for ( int j=1; j < 3; j++ ) {
    const Point3 &v = vertPositions[ vertIndices[v0Index+j] ];
    for ( int k=0; k < 3; k++ ) {
      if ( v[k] < minP[k] )
        minP[k] = v[k];
      if ( v[k] > maxP[k] )
        maxP[k] = v[k];
    }
  }

  // return bounding box of triangle
  return AABB(minP, maxP);
}


bool TriangleMesh::intersectRayTriangle(const Ray &ray,
                                        const Point3 *triV,
                                        const Vector3 &triN,
                                        unsigned char axis,
                                        const Tri &triAttr,
                                        float &distance) {
  // backface test was moved out of this method since it is faster not to call
  // this at all if the triangle is a backface

  // get major axis of triangle normal
  // (stored with triangle so it does not have to be recalculated)
  unsigned char k = axis & 0x03;

  // get the other two axis
  unsigned char u = (axis & 0x0C) >> 2,
                v = (axis & 0x30) >> 4;

  const float lnd = 1.0f/(ray.getDir()[k] +
                          triAttr.nu*ray.getDir()[u] +
                          triAttr.nv*ray.getDir()[v]);
  const float t   = (triAttr.nd -
                     ray.getOrigin()[k] -
                     triAttr.nu*ray.getOrigin()[u] -
                     triAttr.nv*ray.getOrigin()[v])*lnd;

  // TODO: I should also take an early out if t > distance.  This will only
  // work if distance that is passed in is guaranteed to be the distance
  // to the nearest intersection so far.
  if ( t < 0.0f )
    return false;

  float hu = ray.getOrigin()[u] + t*ray.getDir()[u] - triV[0][u],
        hv = ray.getOrigin()[v] + t*ray.getDir()[v] - triV[0][v];

  float beta = hv*triAttr.bnu + hu*triAttr.bnv;
  if ( beta < 0.0f )
    return false;

  float gamma = hu*triAttr.cnu + hv*triAttr.cnv;
  if ( gamma < 0.0f )
    return false;

  if ( beta + gamma > 1.0f )
    return false;

  distance = t;

  return true;
}
