
#ifndef TRIANGLEMESH_H
#define TRIANGLEMESH_H

#include <vector>
#include "SceneObject.h"
#include "Math3D/intersect.h"


bool intersectRayAABB(const Ray &ray, const AABB &aabb,
                      float *tNearOut=NULL, float *tFarOut=NULL);

class TriangleMesh {
public:
  TriangleMesh(int _matID, const char *filename, float scale=1.0f);
  virtual ~TriangleMesh();

  virtual bool intersectRay(const Ray &ray, HitInfo &hitInfo) const=0;
  
  virtual bool occlusionTest(const Ray &ray, float maxT) const = 0;

  void buildAABB();
  
  void computeBaryCentricCoords(int triIndex,
                              const Point3 &p,
                              float b[]) const;

  // this method indicates whether vertex tangents have been
  // computed for this mesh
  inline bool hasVertTangents() const { return vertTangents != NULL; }
  
  // some accessor methods
  inline int            getNumTris()           const { return numTris;           }
  inline const int*     getVertIndices()       const { return vertIndices;       }
  inline const AABB&    getAABB()              const { return aabb;              }
  inline const Point2*  getTexCoords()         const { return vertTexCoords;     }
  inline const Vector3* getVertNormals()       const { return vertNormals;       }
  inline const Vector3* getVertTangents()      const { return vertTangents;      }
  inline const Vector3* getVertBitangents()    const { return vertBitangents;    }
  inline const Vector3* getSurfaceNormals()    const { return surfaceNormals;    }
  inline const Vector3* getSurfaceTangents()   const { return surfaceTangents;   }
  inline const Vector3* getSurfaceBitangents() const { return surfaceBitangents; }
protected:
  struct Tri {
    // TODO: move faces and dominant axis into this structure
    float bnu, bnv;
    float cnu, cnv;
    float nu, nv, nd;
  };

  static bool intersectRayTriangle(const Ray &ray,
                                   const Point3 *triV,
                                   const Vector3 &triN,
                                   unsigned char axis,
                                   const Tri &triAttr,
                                   float &distance);
  void clear();
  void allocateStorage(int newNumTris,
                       int newNumVerts,
                       bool allocTC,
                       bool allocVertNormals);
  void load(const char *filename, float scale=1.0f);
  void load3DS(const char *filename, float scale=1.0f);
  void loadMSH(const char *filename, float scale=1.0f);
  void processTriangle(int i);
  void computeVertexNormals(int numVerts);

  AABB calcTriAABB(int i);

  int matID;
  AABB aabb;
  
  int numTris;                // specifies how many triangles are in the mesh
  int numVertices;            // specifies how many vertices are in the mesh

  int *vertIndices;           // indices of each triangles vertices

  // the following members are parallel arrays (per triangle) that contain information
  // for each triangle.
  Vector3 *surfaceTangents,   // tangent vector to face of triangle
          *surfaceBitangents, // binormal vector to face of triangle
          *surfaceNormals;    // normal vector to face of triangle
  unsigned char *triAxis;     // dominant axis of triangle

  // stores triangle attributes (later should store all per-triangle attributes)
  Tri *triAttrs;

  // the following members are parallel arrays (per vertex) that contain the
  // information for each vertex
  Point3  *vertPositions;  // position of vertex
  Point2  *vertTexCoords;  // texture coordinates
  Vector3 *vertNormals;    // normals at each vertex
  Vector3 *vertBitangents; // bitangents at each vertex
  Vector3 *vertTangents;   // tangents at each vertex
}; // TriangleMesh


#endif
