
#include <algorithm> // for std::sort
#include "TriangleMeshKDTree.h"

// The KD-Tree depth can NEVER exceed this amount. In practical
// cases the depth of the tree will be much less.
#define TRIANGLE_MESH_KDTREE_MAX_DEPTH   64


const float TriangleMeshKDTree::intersectionCost = 80.0f;
const float TriangleMeshKDTree::traversalCost    = 1.0f;
const float TriangleMeshKDTree::emptyBonus       = 0.5f;
const int   TriangleMeshKDTree::maxTrisInLeaf    = 4; // TODO: try other values
const int   TriangleMeshKDTree::maxBadRefines    = 3;


TriangleMeshKDTree::TriangleMeshKDTree(int _matID,
                                       const char *filename,
                                       float scale):
     TriangleMesh(_matID, filename, scale),
     root(NULL) {
  buildKDTree();
}


TriangleMeshKDTree::~TriangleMeshKDTree() {
  if ( root )
    delete root;
}


void TriangleMeshKDTree::buildKDTree() {
  // delete tree if it already exists
  if ( root ) {
    delete root;
    root = NULL;
  }

  // don't continue if there are no triangles
  if ( 0 == numTris )
    return;

  // calculate maximum depth of tree
  int maxTreeDepth = int( 8.0 + 1.3*log( double(numTris) ) );

  // make sure maxTreeDepth is not too big
  if ( maxTreeDepth > TRIANGLE_MESH_KDTREE_MAX_DEPTH )
    maxTreeDepth = TRIANGLE_MESH_KDTREE_MAX_DEPTH;

  // create vector containing all triangle indices in root, this is required
  // by the KDTreeNode constructor which recursively builds the KD-Tree.
  // KDTreeNode() will internally create new vectors of triangle indices
  // to indicate which objects intersect each node.
  std::vector<int> rootTriIndices(numTris);
  for ( int i=0; i < numTris; i++ )
    rootTriIndices[i] = i;

  // store AABB of each triangle so they don't have to be recomputed during
  // the construction of the KD-Tree
  std::vector<AABB> triAABBs(numTris);
  for ( int i=0; i < numTris; i++ )
    triAABBs[i] = calcTriAABB(i);

  // call KDTreeNode contructor which will recursively build the KD-Tree
  root = new KDTreeNode(aabb,
                        vertPositions,
                        vertIndices,
                        maxTrisInLeaf,
                        rootTriIndices,
                        triAABBs,
                        0,
                        maxTreeDepth);

  if ( !root )
    throw RTException("TriangleMeshKDTree::buildKDTree(): unable to allocate root");
}


bool TriangleMeshKDTree::intersectRay(const Ray &ray,
                                      HitInfo &hitInfo) const {
  struct KDTreeNodeQueueEl {
    KDTreeNode *node;
    float tmin, tmax;
  };
  KDTreeNodeQueueEl nodeQueue[TRIANGLE_MESH_KDTREE_MAX_DEPTH];
  int nodeQueuePos = 0;

  // pbrt: compute initial parametric range of ray inside kd-tree extent
  float tmin = 0.0f,
        tmax = 0.0f;
  if ( !intersectRayAABB(ray, aabb, &tmin, &tmax) )
    return false; // ray does not intersect bounding box of scene
  
  // store inverse of ray direction so multiplies can be used instead of
  // divides in loop
  const Vector3 &rayDir    = ray.getDir();
  const Point3  &rayOrigin = ray.getOrigin();
  Vector3 invDir(1.0f/rayDir.m[0],
                 1.0f/rayDir.m[1],
                 1.0f/rayDir.m[2]);

  int nearestHit = -1;
  float nearestHitT = RT_FLT_MAX;

  // pbrt: Traverse kd-tree nodes in order for ray
  KDTreeNode *node = root;
  while ( node != NULL ) {
    // pbrt: bail out if we found a hit closer than the current node
    if ( nearestHitT < tmin )
      break;

    unsigned char axis = node->getAttrs();
    
    if ( axis != 0x03 ) {
      // pbrt: process kd-tree interior node
      const KDTreeNode::InternalNodeData &internalNodeData = node->getInternalNodeData();

      // pbrt: compute parametric distance along ray to split plane
      float tplane = (internalNodeData.planeOffset - rayOrigin.m[axis])*invDir[axis];

      // pbrt: get node children pointers for ray
      KDTreeNode *firstChild,
                 *secondChild;
      if ( rayOrigin.m[axis] <= internalNodeData.planeOffset ) {
        firstChild  = internalNodeData.left;
        secondChild = internalNodeData.right;
      }
      else {
        firstChild  = internalNodeData.right;
        secondChild = internalNodeData.left;
      }

      // pbrt: advance to next child node, possibly enqueue other child
      if ( tplane > tmax || tplane < 0.0f )
        node = firstChild;
      else if ( tplane < tmin )
        node = secondChild;
      else {
        // pbrt: enqueue second child in todo list
        nodeQueue[nodeQueuePos].node = secondChild;
        nodeQueue[nodeQueuePos].tmin = tplane;
        nodeQueue[nodeQueuePos].tmax = tmax;
        nodeQueuePos++;
        node = firstChild;
        tmax = tplane;
      }
    } // if ( internal node )
    else {
      // pbrt: check for intersections inside leaf node
      const KDTreeNode::LeafNodeData &leafNodeData = node->getLeafNodeData();
      for ( int i=0; i < leafNodeData.numTris; i++ ) {
        // test agaisnt this triangle if it is not a backface to this ray
        int triIndex = leafNodeData.tris[i];
        if ( ray.getDir()*surfaceNormals[triIndex] < 0.0f ) {
          float _t = RT_FLT_MAX;
          int v0Index = triIndex*3;
          Point3 triV[3] = { vertPositions[ vertIndices[v0Index  ] ],
                             vertPositions[ vertIndices[v0Index+1] ],
                             vertPositions[ vertIndices[v0Index+2] ] };
          bool result = TriangleMesh::intersectRayTriangle(ray,
                                                           triV,
                                                           surfaceNormals[triIndex],
                                                           triAxis[triIndex],
                                                           triAttrs[triIndex],
                                                           _t);
          if ( result && _t < nearestHitT ) {
            nearestHit = triIndex;
            nearestHitT = _t;
          }
        } // if (triangle is not a backface)
      } // for (triangles in leaf)

      // pbrt: grab next node to process from todo list
      if ( nodeQueuePos > 0 ) {
        nodeQueuePos--;
        node = nodeQueue[nodeQueuePos].node;
        tmin = nodeQueue[nodeQueuePos].tmin;
        tmax = nodeQueue[nodeQueuePos].tmax;
      }
      else
        break;
    } // else ( leaf node )
  } // while ( node != NULL )

  if ( nearestHit >= 0 ) {
    hitInfo.t       = nearestHitT;
    hitInfo.texInfo = nearestHit;
    return true;
  }

  return false;
}


// returns after the first hit within distance maxT is found
bool TriangleMeshKDTree::occlusionTest(const Ray &ray,
                                       float maxT) const {
  struct KDTreeNodeQueueEl {
    KDTreeNode *node;
    float tmin, tmax;
  };
  KDTreeNodeQueueEl nodeQueue[TRIANGLE_MESH_KDTREE_MAX_DEPTH];
  int nodeQueuePos = 0;

  // pbrt: compute initial parametric range of ray inside kd-tree extent
  float tmin = 0.0f,
        tmax = 0.0f;
  if ( !intersectRayAABB(ray, aabb, &tmin, &tmax) )
    return false; // ray does not intersect bounding box of scene

  // store inverse of ray direction so multiplies can be used instead of
  // divides in loop
  const Vector3 &rayDir    = ray.getDir();
  const Point3  &rayOrigin = ray.getOrigin();
  Vector3 invDir(1.0f/rayDir.m[0],
                 1.0f/rayDir.m[1],
                 1.0f/rayDir.m[2]);
  float nearestHitT = RT_FLT_MAX;

  // pbrt: Traverse kd-tree nodes in order for ray
  KDTreeNode *node = root;
  while ( node != NULL ) {
    // pbrt: bail out if we found a hit closer than the current node
    if ( nearestHitT < tmin )
      break;

    unsigned char axis = node->getAttrs();
    
    if ( axis != 0x03 ) {
      // pbrt: process kd-tree interior node
      const KDTreeNode::InternalNodeData &internalNodeData = node->getInternalNodeData();

      // pbrt: compute parametric distance along ray to split plane
      float tplane = (internalNodeData.planeOffset - rayOrigin.m[axis])*invDir[axis];

      // pbrt: get node children pointers for ray
      KDTreeNode *firstChild,
                 *secondChild;
      if ( rayOrigin.m[axis] <= internalNodeData.planeOffset ) {
        firstChild  = internalNodeData.left;
        secondChild = internalNodeData.right;
      }
      else {
        firstChild  = internalNodeData.right;
        secondChild = internalNodeData.left;
      }

      // pbrt: advance to next child node, possibly enqueue other child
      if ( tplane > tmax || tplane < 0.0f )
        node = firstChild;
      else if ( tplane < tmin )
        node = secondChild;
      else {
        // pbrt: enqueue second child in todo list
        nodeQueue[nodeQueuePos].node = secondChild;
        nodeQueue[nodeQueuePos].tmin = tplane;
        nodeQueue[nodeQueuePos].tmax = tmax;
        nodeQueuePos++;
        node = firstChild;
        tmax = tplane;
      }
    } // if ( internal node )
    else {
      // pbrt: check for intersections inside leaf node
      const KDTreeNode::LeafNodeData &leafNodeData = node->getLeafNodeData();
      for ( int i=0; i < leafNodeData.numTris; i++ ) {
        // test agaisnt this triangle if it is not a backface to this ray
        int triIndex = leafNodeData.tris[i];
        if ( ray.getDir()*surfaceNormals[triIndex] < 0.0f ) {
          float _t = RT_FLT_MAX;
          int v0Index = triIndex*3;
          Point3 triV[3] = { vertPositions[ vertIndices[v0Index  ] ],
                             vertPositions[ vertIndices[v0Index+1] ],
                             vertPositions[ vertIndices[v0Index+2] ] };
          if ( TriangleMesh::intersectRayTriangle(ray,
                                                  triV,
                                                  surfaceNormals[triIndex],
                                                  triAxis[triIndex],
                                                  triAttrs[triIndex],
                                                  _t) ) {
            if ( _t < maxT )
              return true; // ray hits this object within distance given
                           // by maxT
            
            // ray hits triangle, but hit is not within given distance.
            
            // check if this is the closest hit so far (so we don't
            // traverse KD-Tree beyond this intersection point)
            if ( _t < nearestHitT )
              nearestHitT = _t;
          } // if (ray hits triangle)
        } // if (triangle is not a backface)
      } // for (triangles in node)

      // pbrt: grab next node to process from todo list
      if ( nodeQueuePos > 0 ) {
        nodeQueuePos--;
        node = nodeQueue[nodeQueuePos].node;
        tmin = nodeQueue[nodeQueuePos].tmin;
        tmax = nodeQueue[nodeQueuePos].tmax;
      }
      else
        break;
    } // else ( leaf node )
  } // while ( node != NULL )

  return false; // ray does not hit object within distance given by maxT
}


TriangleMeshKDTree::KDTreeNode::KDTreeNode(const AABB &nodeAABB,
                                           const Point3 *triVerts,
                                           const int *triIndices,
                                           int maxLeafTris,
                                           const std::vector<int> &trisInNode,
                                           const std::vector<AABB> &triAABBs,
                                           int badRefines,
                                           int depth) {
  // zero out self before doing anything
  // just memset the larger of the two node structures to 0, and
  // set attrs to zero
  attrs = 0;
  memset( &internalNodeData, 0, sizeof(internalNodeData) );

  // termination criteria met?
  if ( int(trisInNode.size()) < maxLeafTris || depth < 1 ) {
    // create a leaf node
    createLeafNode(trisInNode);
  }
  else {
    // try to create an internal node
    createInternalNode(nodeAABB,
                       triVerts,
                       triIndices,
                       trisInNode,
                       triAABBs,
                       maxLeafTris,
                       badRefines,
                       depth);
  }
}


TriangleMeshKDTree::KDTreeNode::~KDTreeNode() {
  // is this an internal node, or a leaf node?
  if ( isLeaf() ) {
    // leaf node
    if ( leafNodeData.tris )
      delete [] leafNodeData.tris;
  }
  else {
    // internal node
    if ( internalNodeData.left )
      delete internalNodeData.left;
    if ( internalNodeData.right )
      delete internalNodeData.right;
  }
}


void TriangleMeshKDTree::KDTreeNode::createLeafNode(const std::vector<int> &trisInNode) {
  // set attrs to indicate this is a leaf
  attrs = 0x03;

  leafNodeData.numTris = int( trisInNode.size() );
  leafNodeData.tris = new int[leafNodeData.numTris];

  // throw an error if could not allocate storage for objects
  if ( !leafNodeData.tris )
    throw RTException("TriangleMeshKDTree::KDTreeNode::createLeaf(): unable to allocate storage for indices");

  for ( int i=0; i < leafNodeData.numTris; i++ )
    leafNodeData.tris[i] = trisInNode[i];
}


void TriangleMeshKDTree::KDTreeNode::createInternalNode(const AABB &nodeAABB,
                                                        const Point3 *triVerts,
                                                        const int *triIndices,
                                                        const std::vector<int> &trisInNode,
                                                        const std::vector<AABB> &triAABBs,
                                                        int maxLeafTris,
                                                        int badRefines,
                                                        int depth) {
  // choose split axis position for interior node
  unsigned char bestAxis = 3;
  size_t bestOffsetIndex = 0;
  float oldCost  = float( trisInNode.size() )*intersectionCost,
        bestCost = RT_FLT_MAX;
        //bestCost = INFINITY;
  Vector3 d = nodeAABB.getMax() - nodeAABB.getMin();
  float totalSA = 2.0f*(d[0]*d[1] + d[0]*d[2] + d[1]*d[2]);
  float invTotalSA = 1.0f/totalSA;

  // choose which axis to split along
  unsigned char axis = d[0] > d[1] ? 0 : 1;
  if ( d[2] > d[axis] )
    axis = 2;

  std::vector<BoxEdge> boxEdges[3];

  int retries = 0;
  float nodeAABBMin,
        nodeAABBMax;
  retrySplit:

  boxEdges[axis].resize(trisInNode.size()*2);
  for ( size_t i=0; i < trisInNode.size(); i++ ) {
    int triIndex = trisInNode[i];
    boxEdges[axis][2*i  ] = BoxEdge(triAABBs[triIndex].getMin().m[axis], triIndex, true);
    boxEdges[axis][2*i+1] = BoxEdge(triAABBs[triIndex].getMax().m[axis], triIndex, false);
  }

  // sort the edges
  std::sort( boxEdges[axis].begin(), boxEdges[axis].end() );

  // compute cost of all potential splits for axis to find best
  nodeAABBMin = nodeAABB.getMin()[axis],
  nodeAABBMax = nodeAABB.getMax()[axis];
  size_t numBelow = 0,
         numAbove = trisInNode.size();
  for ( size_t i=0; i < 2*trisInNode.size(); i++ ) {
    const BoxEdge &edge = boxEdges[axis][i];
    if ( false == edge.start )
      numAbove--;
    if ( edge.pos > nodeAABBMin &&
         edge.pos < nodeAABBMax ) {
      // compute cost for split at ith edge
      static const int otherAxis[3][2] = { {1, 2}, {0, 2}, {0, 1} };
      int otherAxis0 = otherAxis[axis][0],
          otherAxis1 = otherAxis[axis][1];

      // calculate surface area above and below the split
      float belowSA = 2.0f*( d[otherAxis0]*d[otherAxis1] +
                             (edge.pos - nodeAABBMin)*
                             (d[otherAxis0] + d[otherAxis1]) ),
            aboveSA = 2.0f*( d[otherAxis0]*d[otherAxis1] +
                             (nodeAABBMax - edge.pos)*
                             (d[otherAxis0] + d[otherAxis1]) );

      // compute probabilities that a random ray hitting this node will cross
      // above or below the split
      float pBelow = belowSA*invTotalSA,
            pAbove = aboveSA*invTotalSA;

      // add extra weight if empty above or below split
      float eb = ( numAbove == 0 || numBelow == 0 ) ? emptyBonus : 0.0f;

      float cost = traversalCost + intersectionCost*(1.0f - eb)*
                   (pBelow*float(numBelow) + pAbove*float(numAbove));

      // update best split if this is lowest cost so far
      if ( cost < bestCost ) {
        bestCost = cost;
        bestAxis = axis;
        bestOffsetIndex = i;
      }
    }
    if ( true == edge.start )
      numBelow++;
  } // for (edges)

  // create leaf if no good splits were found
  if ( 3 == bestAxis && retries < 2 ) {
    retries++;
    axis = (axis+1)%3; // try next axis
    goto retrySplit;
  }

  if ( bestCost > oldCost )
    badRefines++;

  if ( (bestCost > 4.0f*oldCost && trisInNode.size() < 16) ||
       3 == bestAxis || maxBadRefines == badRefines ) {
    // create a leaf
    createLeafNode(trisInNode);
    return; // don't go any further after creating leaf
  }

  // classify primitives with respect to split
  std::vector<int> tris0;
  for ( size_t i=0; i < bestOffsetIndex; i++ ) {
    const BoxEdge &edge = boxEdges[bestAxis][i];
    if ( true == edge.start )
      tris0.push_back(edge.triIndex);
  }

  std::vector<int> tris1;
  for ( size_t i=bestOffsetIndex+1; i < 2*trisInNode.size(); i++ ) {
    const BoxEdge &edge = boxEdges[bestAxis][i];
    if ( false == edge.start )
      tris1.push_back(edge.triIndex);
  }

  // recursively initialize child nodes
  Point3 box1Min = nodeAABB.getMin(),
         box0Max = nodeAABB.getMax();
  box0Max.m[bestAxis] = box1Min.m[bestAxis] = boxEdges[bestAxis][bestOffsetIndex].pos;
  AABB box0( nodeAABB.getMin(), box0Max ),
       box1( box1Min, nodeAABB.getMax() );

  // set attrs to best split axis to indicate that this is an internal node
  attrs = bestAxis;

  // store offset of splitting plane
  internalNodeData.planeOffset = boxEdges[bestAxis][bestOffsetIndex].pos;

  // create child nodes
  internalNodeData.left = new KDTreeNode(box0,
                                         triVerts,
                                         triIndices,
                                         maxLeafTris,
                                         tris0,
                                         triAABBs,
                                         badRefines,
                                         depth-1);

  if ( !internalNodeData.left )
    throw RTException("TriangleMeshKDTree::KDTreeNode::createInternalNode(): "
                      "unable to create child node");

  internalNodeData.right = new KDTreeNode(box1,
                                          triVerts,
                                          triIndices,
                                          maxLeafTris,
                                          tris1,
                                          triAABBs,
                                          badRefines,
                                          depth-1);

  if ( !internalNodeData.right )
    throw RTException("TriangleMeshKDTree::KDTreeNode::createInternalNode(): "
                      "unable to create child node");
}



TriangleMeshKDTree::BoxEdge::BoxEdge(float _pos, int _triIndex, bool _start):
     pos(_pos),
     triIndex(_triIndex),
     start(_start) {

}



bool TriangleMeshKDTree::BoxEdge::operator < (const BoxEdge &e) const {
  if ( pos == e.pos )
    return (true == start && false == e.start);

  return pos < e.pos;
}
