
#include "TriangleMeshObject.h"


TriangleMeshObject::TriangleMeshObject(const TriangleMesh &_mesh,
                                       int _matID,
                                       const Point3 &_pos,
                                       float rotX,
                                       float rotY,
                                       float rotZ):
     mesh(_mesh),
     matID(_matID),
     pos(_pos) {
  AABB meshAABB = mesh.getAABB();
  aabb = AABB( pos + meshAABB.getCenter(), meshAABB.getExt() );
  setOrientation(rotX, rotY, rotZ);
}


TriangleMeshObject::~TriangleMeshObject() {
  // nothing to clean up
}


void TriangleMeshObject::getNormalAndTCAt(const Point3 &p,
                                          int texInfo,
                                          const Material &mat,
                                          Point2 &tc,
                                          Vector3 &normal) const {
  // don't continue if index of hit triangle is not set (or is invalid)
    if ( texInfo < 0 ||
       texInfo >= mesh.getNumTris() ) {
    tc.m[0] = tc.m[1] = 0.0f;
    return;  
  }
  
  // get pointers to triangle data
  const Point2  *vertTexCoords = mesh.getTexCoords();
  const int     *vertIndices   = mesh.getVertIndices();
  const Vector3 *vertNormals   = mesh.getVertNormals();
  
  // transform hitpoint into this object's space
  Point3 transformedHit = invRotMat*(p - pos);

  // compute barycentric coordinates  
  float b[3] = { 0.0f, 0.0f, 0.0f };
  mesh.computeBaryCentricCoords(texInfo, transformedHit, b);

  // get indices of vertices (they will be needed in code that follows)  
  int v0Index = texInfo*3;
  int v0 = vertIndices[v0Index  ],
      v1 = vertIndices[v0Index+1],
      v2 = vertIndices[v0Index+2];
  
  // compute texture coordinates
  tc[0] = tc[1] = 0.0f;
  if ( vertTexCoords ) {
    // material has a texture, so compute texture coordinates
    Point2 vertTC0 = vertTexCoords[v0],
           vertTC1 = vertTexCoords[v1],
           vertTC2 = vertTexCoords[v2];
    tc[0] = b[0]*vertTC0.m[0] + b[1]*vertTC1.m[0] + b[2]*vertTC2.m[0];
    tc[1] = b[0]*vertTC0.m[1] + b[1]*vertTC1.m[1] + b[2]*vertTC2.m[1];
  }
  
  // does this material have a normal map?
  if ( mat.normalMap ) {
    // no normal mapping in this iteration of the project
  }
  else {
    // material does not have a normal map
    
    // if there are vertex normals then interpolate between them
    if ( vertNormals ) {
      const Vector3 &n0 = vertNormals[v0],
                    &n1 = vertNormals[v1],
                    &n2 = vertNormals[v2];
      normal = n0*b[0] + n1*b[1] + n2*b[2];
      normal.normalize();
    }
    else {
      // no vertex normals, so fall back to surface normal
      // (surace normals should always be set, and they should
      //  be normalized)
      normal = mesh.getSurfaceNormals()[texInfo];
    }
  }
}


void TriangleMeshObject::setOrientation(float rotX,
                                        float rotY,
                                        float rotZ) {
  // compute quaternion from Euler angles
  orient = Quat::fromEulerAngles(rotX, rotY, rotZ);
  
  // compute inverse rotation matrix
  // this is needed to transform rays into this object's space
  // store it so it does not need to be computed on the fly
  invRotMat = orient.inverse().toMatrix3();
}


bool TriangleMeshObject::intersectRay(const Ray &ray,
                                      HitInfo &hitInfo) const {
  // transform ray into this object's frame of reference
  Ray transformedRay;
  transformedRay.setOrigin( invRotMat*(ray.getOrigin() - pos) );
  transformedRay.setDirDontNormalize( invRotMat*ray.getDir() );

  return mesh.intersectRay(transformedRay, hitInfo);;
}


bool TriangleMeshObject::occlusionTest(const Ray &ray,
                                       float maxT) const {
  // transform ray into this object's frame of reference
  Ray transformedRay;
  transformedRay.setOrigin( invRotMat*(ray.getOrigin() - pos) );
  transformedRay.setDirDontNormalize( invRotMat*ray.getDir() );

  return mesh.occlusionTest(transformedRay, maxT);
}


AABB TriangleMeshObject::getAABB() const {
  // TODO: rotate extents to account for orientation of object
  return aabb;
}


void TriangleMeshObject::setPos(const Point3 &p) {
  pos = p;
  aabb = AABB( pos, aabb.getExt() );
}

